@@ -23,7 +23,6 @@ from __future__ import (
 
 import json
 import logging
-import os
 import sys
 import time
 
@@ -31,13 +30,14 @@ import jujuclient
 
 from quickstart import (
     juju,
+    netutils,
     platform_support,
     settings,
     ssh,
     utils,
     watchers,
 )
-from quickstart.models import envs
+from quickstart.models import jenv
 
 
 class ProgramExit(Exception):
@@ -116,6 +116,19 @@ def ensure_dependencies(distro_only, platform, juju_command):
     return juju_version
 
 
+def check_juju_supported(juju_version):
+    """Ensure the given Juju version is supported by Quickstart.
+
+    Raise a ProgramExit if Juju is not supported.
+    """
+    if juju_version < settings.JUJU_SUPPORTED_VERSION:
+        current = b'.'.join(map(bytes, juju_version))
+        supported = b'.'.join(map(bytes, settings.JUJU_SUPPORTED_VERSION))
+        raise ProgramExit(
+            b'the current Juju version ({}) is not supported: '
+            b'please upgrade to Juju >= {}'.format(current, supported))
+
+
 def ensure_ssh_keys():
     """Ensure that SSH keys are available.
 
@@ -163,30 +176,56 @@ def ensure_ssh_keys():
         raise ProgramExit(bytes(err))
 
 
-def bootstrap(env_name, juju_command, requires_sudo=False, debug=False,
-              upload_tools=False, upload_series=None, constraints=None):
+def check_bootstrapped(env_name):
+    """Check if the environment named env_name is already bootstrapped.
+
+    If so, return the environment API URL to be used to connect to the Juju API
+    server. If not already bootstrapped, or if the API URL cannot be retrieved,
+    return None.
+    """
+    if not jenv.exists(env_name):
+        return None
+    # Retrieve the Juju API addresses from the jenv file.
+    try:
+        candidates = jenv.get_value(env_name, 'state-servers')
+    except ValueError as err:
+        logging.warn(b'cannot retrieve the Juju API URL: {}'.format(err))
+        return None
+    # Look for a reachable API URL.
+    if not candidates:
+        logging.warn('cannot retrieve the Juju API URL: no addresses found')
+        return None
+    for candidate in candidates:
+        error = netutils.check_listening(candidate)
+        if error is None:
+            # Juju API URL found.
+            return 'wss://{}'.format(candidate)
+        logging.debug(error)
+    logging.warn(
+        'cannot retrieve the Juju API URL: cannot connect to any of the '
+        'following addresses: {}'.format(', '.join(candidates)))
+    return None
+
+
+def bootstrap(
+        env_name, juju_command, debug=False, upload_tools=False,
+        upload_series=None, constraints=None):
     """Bootstrap the Juju environment with the given name.
 
+    Return a flag indicating whether the environment was already bootstrapped.
+
     Do not bootstrap the environment if already bootstrapped.
+    If the environment is not bootstrapped, execute the bootstrap command with
+    the given juju_command, debug, upload_tools, upload_series and constraints
+    arguments.
 
-    Return a tuple (already_bootstrapped, series) in which:
-        - already_bootstrapped indicates whether the environment was already
-          bootstrapped;
-        - series is the bootstrap node Ubuntu series.
-
-    The is_local argument indicates whether the environment is configured to
-    use the local provider. If so, sudo privileges are requested in order to
-    bootstrap the environment.
-
-    If debug is True and the environment not bootstrapped, execute the
-    bootstrap command passing the --debug flag.
+    When the function exists the Juju environment is bootstrapped, but we don't
+    know if it is ready yet. For this reason, a call to status() is usually
+    required at that point.
 
     Raise a ProgramExit if any error occurs in the bootstrap process.
     """
-    already_bootstrapped = False
     cmd = [juju_command, 'bootstrap', '-e', env_name]
-    if requires_sudo:
-        cmd.insert(0, 'sudo')
     if debug:
         cmd.append('--debug')
     if upload_tools:
@@ -196,34 +235,28 @@ def bootstrap(env_name, juju_command, requires_sudo=False, debug=False,
     if constraints is not None:
         cmd.extend(['--constraints', constraints])
     retcode, _, error = utils.call(*cmd)
-    if retcode:
-        # XXX frankban 2013-11-13 bug 1252322: the check below is weak. We are
-        # relying on an error message in order to decide if the environment is
-        # already bootstrapped. Other possibilities include checking if the
-        # jenv file is present (in ~/.juju/environments/) and, if so, check the
-        # juju status. Unfortunately this is also prone to errors, because a
-        # jenv file can be there but the environment not really bootstrapped or
-        # functional (e.g. sync-tools was used, or a previous bootstrap failed,
-        # or the user terminated machines from the ec2 panel, etc.). Moreover
-        # jenv files seems to be an internal juju-core detail. Definitely we
-        # need to find a better way, but for now the "asking forgiveness"
-        # approach feels like the best compromise we have. Also note that,
-        # rather than comparing the expected error with the obtained one, we
-        # search in the error in order to support bootstrap --debug.
-        if 'environment is already bootstrapped' not in error:
-            # We exit if the error is not "already bootstrapped".
-            raise ProgramExit(error)
-        # Juju is bootstrapped, but we don't know if it is ready yet. Fall
-        # through to the next block for that check.
-        already_bootstrapped = True
-        print('reusing the already bootstrapped {} environment'.format(
-            env_name))
-    # Call "juju status" multiple times until the bootstrap node is ready.
-    # Exit with an error if the agent is not ready after ten minutes.
-    # Note: when using the local provider, calling "juju status" is very fast,
-    # but e.g. on ec2 the first call (right after "bootstrap") can take
-    # several minutes, and subsequent calls are relatively fast (seconds).
-    print('retrieving the environment status')
+    if not retcode:
+        return False
+    # XXX frankban 2013-11-13 bug 1252322: the check below is weak. We are
+    # relying on an error message in order to decide if the environment is
+    # already bootstrapped.  Also note that, rather than comparing the expected
+    # error with the obtained one, we search in the error in order to support
+    # bootstrap --debug.
+    if 'environment is already bootstrapped' not in error:
+        # We exit if the error is not "already bootstrapped".
+        raise ProgramExit(error)
+    # Juju is already bootstrapped.
+    return True
+
+
+def status(env_name, juju_command):
+    """Call "juju status" multiple times until the bootstrap node is ready.
+
+    Return the bootstrap node series of the Juju environment.
+
+    Raise a ProgramExit if the agent is not ready after ten minutes or if the
+    agent is in an error state.
+    """
     timeout = time.time() + (60*10)
     while time.time() < timeout:
         retcode, output, error = utils.call(
@@ -236,8 +269,7 @@ def bootstrap(env_name, juju_command, requires_sudo=False, debug=False,
         except ValueError:
             continue
         if agent_state == 'started':
-            series = utils.get_bootstrap_node_series(output)
-            return already_bootstrapped, series
+            return utils.get_bootstrap_node_series(output)
         # If the agent is in an error state, there is nothing we can do, and
         # it's not useful to keep trying.
         if agent_state == 'error':
@@ -246,27 +278,17 @@ def bootstrap(env_name, juju_command, requires_sudo=False, debug=False,
     raise ProgramExit('the state server is not ready:\n{}'.format(details))
 
 
-def get_value_from_jenv(env_name, juju_home, key):
-    """Read the key from the generated environment file.
+def get_credentials(env_name):
+    """Return the Juju credentials for the given environment name.
 
-    At bootstrap, juju (v1.16 and later) writes a generated file
-    located in JUJU_HOME.  Return the value for the key.
-
-    Raise a ValueError if:
-        - the environment file is not found;
-        - the environment file contents are not parsable by YAML;
-        - the YAML contents are not properly structured;
-        - the key is not found.
+    Parse the jenv file to retrieve the environment user name and password.
+    Raise a ProgramExit if the credentials cannot be retrieved.
     """
-    filename = '{}.jenv'.format(env_name)
-    juju_env_file = os.path.expanduser(
-        os.path.join(juju_home, 'environments', filename))
-    jenv_db = envs.load_generated(juju_env_file)
     try:
-        return jenv_db[key]
-    except KeyError:
-        msg = '{} not found in {}'.format(key, juju_env_file)
-        raise ValueError(msg.encode('utf-8'))
+        return jenv.get_credentials(env_name)
+    except ValueError as err:
+        msg = b'cannot retrieve environment credentials: {}'.format(err)
+        raise ProgramExit(msg)
 
 
 def get_api_url(env_name, juju_command):
@@ -286,7 +308,7 @@ def get_api_url(env_name, juju_command):
     return 'wss://{}'.format(api_address)
 
 
-def connect(api_url, admin_secret):
+def connect(api_url, username, password):
     """Connect to the Juju API server and log in using the given credentials.
 
     Return a connected and authenticated Juju Environment instance.
@@ -309,13 +331,26 @@ def connect(api_url, admin_secret):
         else:
             break
     try:
-        env.login(admin_secret)
+        env.authenticate(username, password)
     except jujuclient.EnvError as err:
         msg = 'unable to log in to the Juju API server on {}: {}'
         raise ProgramExit(msg.format(api_url, err.message))
     return env
 
 
+def get_env_type(env):
+    """Return the Juju environment type for the given environment connection.
+
+    Raise a ProgramExit if the environment type cannot be retrieved.
+    """
+    try:
+        info = env.info()
+    except jujuclient.EnvError as err:
+        msg = 'cannot retrieve the environment type: {}'.format(err.message)
+        raise ProgramExit(msg)
+    return info['ProviderType']
+
+
 def create_auth_token(env):
     """Return a new authentication token.
 
@@ -351,8 +386,8 @@ def check_environment(
           unit.
 
     If the charm URL is not provided, and the service is not already deployed,
-    the function tries to retrieve it from the charmworld API. In this case a
-    default charm URL is used if charmworld is not available.
+    the function tries to retrieve it from the charm store API. In this case a
+    default charm URL is used if the charm store service is not available.
 
     Return a tuple including the following values:
         - charm_url: the charm URL that will be used to deploy the service;
@@ -383,8 +418,8 @@ def check_environment(
             else:
                 series = settings.JUJU_GUI_SUPPORTED_SERIES[-1]
             try:
-                # Try to get the charm URL from charmworld.
-                charm_url = utils.get_charm_url(series)
+                # Try to get the charm URL from the charm store API.
+                charm_url = netutils.get_charm_url(series)
             except (IOError, ValueError) as err:
                 # Fall back to the default URL for the current series.
                 msg = 'unable to retrieve the {} charm URL from the API: {}'
@@ -497,14 +532,10 @@ def watch(env, unit_name):
         for action, data in unit_changes:
             if data['Name'] == unit_name:
                 try:
-                    data = watchers.parse_unit_change(
-                        action, data, unit_status, address)
+                    unit_status, machine_id = watchers.parse_unit_change(
+                        action, data, unit_status)
                 except ValueError as err:
                     raise ProgramExit(bytes(err))
-                unit_status, address, machine_id = data
-                if address and unit_status == 'started':
-                    # We can exit this loop.
-                    return address
                 # The mega-watcher contains a single change for each specific
                 # unit. For this reason, we can exit the for loop here.
                 break
@@ -528,12 +559,12 @@ def watch(env, unit_name):
                         action, data, machine_status, address)
                 except ValueError as err:
                     raise ProgramExit(bytes(err))
-                if address and unit_status == 'started':
-                    # We can exit this loop.
-                    return address
                 # The mega-watcher contains a single change for each specific
                 # machine. For this reason, we can exit the for loop here.
                 break
+        if address and unit_status == 'started':
+            # We can exit this loop.
+            return address
 
 
 def deploy_bundle(env, bundle_yaml, bundle_name, bundle_id):
