@@ -94,7 +94,10 @@ views, including the App object, are properly provided. This is effectively the
 proposed solution to build multi-views CLI applications in Quickstart.
 """
 
-from __future__ import unicode_literals
+from __future__ import (
+    print_function,
+    unicode_literals,
+)
 
 import copy
 import functools
@@ -103,6 +106,7 @@ import operator
 import urwid
 
 from quickstart import (
+    maas,
     platform_support,
     settings,
 )
@@ -111,7 +115,10 @@ from quickstart.cli import (
     forms,
     ui,
 )
-from quickstart.models import envs
+from quickstart.models import (
+    envs,
+    jenv,
+)
 
 
 def show(view, *args):
@@ -130,49 +137,53 @@ def show(view, *args):
         # Start the Urwid interactive session (main loop).
         loop.run()
     except ui.AppExit as err:
+        # The print below ensures the screen is properly refreshed when
+        # exiting the interactive session.
+        print('interactive session closed')
         return err.return_value
 
 
-def env_index(app, env_type_db, env_db, save_callable):
+def _save_and_exit(env_db, env_data, save_callable):
+    """Add the new environment env_data to the env_db.
+
+    Exit the interactive session passing the newly saved environment, so that
+    it is bootstrapped by the application.
+    """
+    envs.set_env_data(env_db, None, env_data)
+    save_callable(env_db)
+    # Use the newly created environment.
+    raise ui.AppExit((env_db, env_data))
+
+
+def env_index(app, params):
     """Show the Juju environments list.
 
     The env_detail view is displayed when the user clicks on an environment.
     From here it is also possible to switch to the edit view in order to create
     a new environment.
 
-    Receives:
+    Receives a params namedtuple-like object including the following fields:
         - env_type_db: the environments meta information;
         - env_db: the environments database;
+        - jenv_db: the jenv files database;
         - save_callable: a function called to save a new environment database.
+    See quickstart.cli.params.Params.
     """
-    env_db = copy.deepcopy(env_db)
+    # XXX frankban 16/12/2014: this function is too long, subdivide it.
+    params = params.copy()
     # All the environment views return a tuple (new_env_db, env_data).
     # Set the env_data to None in the case the user quits the application
     # without selecting an environment to use.
-    app.set_return_value_on_exit((env_db, None))
-    detail_view = functools.partial(
-        env_detail, app, env_type_db, env_db, save_callable)
-    edit_view = functools.partial(
-        env_edit, app, env_type_db, env_db, save_callable)
+    app.set_return_value_on_exit((params.env_db, None))
+    detail_view = functools.partial(env_detail, app, params)
+    jenv_view = functools.partial(jenv_detail, app, params)
+    edit_view = functools.partial(env_edit, app, params)
     # Alphabetically sort the existing environments.
     environments = sorted([
-        envs.get_env_data(env_db, env_name)
-        for env_name in env_db['environments']
+        envs.get_env_data(params.env_db, env_name)
+        for env_name in params.env_db['environments']
     ], key=operator.itemgetter('name'))
 
-    def create_and_start_local_env():
-        # Automatically create and use a local environment named "local".
-        # This closure can only be called when there are no environments in the
-        # database. For this reason, the new environment is set as default.
-        # Exit the interactive session selecting the newly created environment.
-        env_data = envs.create_local_env_data(
-            env_type_db, 'local', is_default=True)
-        # Add the new environment to the environments database.
-        envs.set_env_data(env_db, None, env_data)
-        save_callable(env_db)
-        # Use the newly created environment.
-        raise ui.AppExit((env_db, env_data))
-
     platform = platform_support.get_platform()
     supports_local = platform_support.supports_local(platform)
 
@@ -196,20 +207,50 @@ def env_index(app, env_type_db, env_db, save_callable):
                 ('highlight', '$ juju quickstart -i'),
             ]),
         ]
+        # If the MAAS CLI is available and it is logged in to a remote API,
+        # offer to automatically create and bootstrap a MAAS environment with
+        # the info retrieved by calling the MAAS CLI.
+        if maas.cli_available():
+            maas_api_data = maas.get_api_info()
+            if maas_api_data is not None:
+                maas_name, maas_server, maas_api_key = maas_api_data
+                maas_callback = ui.thunk(
+                    _create_and_start_maas_env,
+                    params.env_type_db, params.env_db, params.save_callable,
+                    maas_name, maas_server, maas_api_key)
+                widgets.extend([
+                    urwid.Text([
+                        '\nThe ',
+                        ('highlight', maas_name),
+                        ' MAAS (bare metal) remote API at ',
+                        ('highlight', maas_server),
+                        ' has been detected, and can be used to '
+                        'automatically set up a MAAS environment. '
+                        'To do that, just click the link below:',
+                    ]),
+                    urwid.Divider(),
+                    ui.MenuButton(
+                        '\N{BULLET} automatically create and bootstrap the {} '
+                        'MAAS environment'.format(maas_name), maas_callback),
+                ])
+
         # If the current platform supports local Juju environments, add an
         # option to automatically create and bootstrap one.
         if supports_local:
+            local_callback = ui.thunk(
+                _create_and_start_local_env,
+                params.env_type_db, params.env_db, params.save_callable)
             widgets.extend([
                 urwid.Text([
                     '\nAt the bottom of the page you can find links to '
                     'manually create new environments. If you instead prefer '
                     'to quickly start your Juju experience in a local '
-                    'environment (LXC), just click the link below:'
+                    'environment (LXC), just click the link below:',
                 ]),
                 urwid.Divider(),
                 ui.MenuButton(
                     '\N{BULLET} automatically create and bootstrap a local '
-                    'environment', ui.thunk(create_and_start_local_env)),
+                    'environment', local_callback),
             ])
 
     app.set_title(title)
@@ -219,8 +260,9 @@ def env_index(app, env_type_db, env_db, save_callable):
     # "juju status" for each environment in the list, which is expensive and
     # time consuming.
     focus_position = None
-    errors_found = default_found = False
+    active_found = default_found = errors_found = False
     existing_widgets_num = len(widgets)
+    remaining_jenv_db = copy.deepcopy(params.jenv_db)
     for position, env_data in enumerate(environments):
         bullet = '\N{BULLET}'
         # Is this environment the default one?
@@ -229,22 +271,51 @@ def env_index(app, env_type_db, env_db, save_callable):
             # The first two positions are the section header and the divider.
             focus_position = position + existing_widgets_num
             bullet = '\N{CHECK MARK}'
-        # Is this environment valid?
-        env_metadata = envs.get_env_metadata(env_type_db, env_data)
-        errors = envs.validate(env_metadata, env_data)
-        if errors:
-            errors_found = True
-            bullet = ('error', bullet)
+        if remaining_jenv_db['environments'].pop(env_data['name'], None):
+            # This is an active environment. Is it running? Who knows...
+            active_found = True
+            bullet = ('active', bullet)
+        else:
+            # Check if this environment is valid.
+            env_metadata = envs.get_env_metadata(params.env_type_db, env_data)
+            errors = envs.validate(env_metadata, env_data)
+            if errors:
+                errors_found = True
+                bullet = ('error', bullet)
         # Create a label for the environment.
         env_short_description = envs.get_env_short_description(env_data)
         text = [bullet, ' {}'.format(env_short_description)]
         widgets.append(ui.MenuButton(text, ui.thunk(detail_view, env_data)))
 
+    # Alphabetically sort the remaining environments not included in the
+    # environments.yaml file.
+    environments = list(sorted([
+        envs.get_env_data(remaining_jenv_db, env_name)
+        for env_name in remaining_jenv_db['environments']
+    ], key=operator.itemgetter('name')))
+
+    if environments:
+        # List the remaining active environments. Those environments are not
+        # included in the environments.yaml file: they are probably imported
+        # and assumed to be working/active. The user has the ability to select
+        # them.
+        widgets.extend([
+            urwid.Divider(),
+            urwid.Text(('highlight', 'Other active environments')),
+            urwid.Text(
+                '(imported/not included in your environments.yaml file):'),
+            urwid.Divider(),
+        ])
+        bullet = ('active', '\N{BULLET}')
+        for env_data in environments:
+            env_short_description = jenv.get_env_short_description(env_data)
+            text = [bullet, ' {}'.format(env_short_description)]
+            widgets.append(ui.MenuButton(text, ui.thunk(jenv_view, env_data)))
+
     # Set up the "create a new environment" section.
     widgets.extend([
         urwid.Divider(),
-        urwid.Text((
-            'highlight', 'Create a new environment:')),
+        urwid.Text(('highlight', 'Create a new environment:')),
         urwid.Divider(),
     ])
     # The Juju GUI can be safely installed in the bootstrap node only if its
@@ -257,7 +328,7 @@ def env_index(app, env_type_db, env_db, save_callable):
     if not supports_local:
         filter_function = lambda env_type, _: env_type != 'local'
     supported_env_types = envs.get_supported_env_types(
-        env_type_db, filter_function=filter_function)
+        params.env_type_db, filter_function=filter_function)
     # Add the buttons used to create new environments.
     widgets.extend([
         ui.MenuButton(
@@ -272,6 +343,8 @@ def env_index(app, env_type_db, env_db, save_callable):
     status = [' \N{UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW} navigate ']
     if default_found:
         status.append(' \N{CHECK MARK} default ')
+    if active_found:
+        status.extend([('active status', ' \N{BULLET}'), ' active '])
     if errors_found:
         status.extend([('error status', ' \N{BULLET}'), ' has errors '])
     app.set_status(status)
@@ -282,63 +355,54 @@ def env_index(app, env_type_db, env_db, save_callable):
     app.set_contents(contents)
 
 
-def env_detail(app, env_type_db, env_db, save_callable, env_data):
+def _create_and_start_local_env(env_type_db, env_db, save_callable):
+    """Automatically create and use a local environment named "local".
+
+    This function can only be called when there are no environments in the
+    database. For this reason, the new environment is set as default.
+    Exit the interactive session selecting the newly created environment.
+    """
+    env_data = envs.create_local_env_data(
+        env_type_db, 'local', is_default=True)
+    _save_and_exit(env_db, env_data, save_callable)
+
+
+def _create_and_start_maas_env(
+        env_type_db, env_db, save_callable, name, server, api_key):
+    """Automatically create and use a MAAS environment.
+
+    This function can only be called when there are no environments in the
+    database. For this reason, the new environment is set as default.
+    Exit the interactive session selecting the newly created environment.
+    """
+    env_data = envs.create_maas_env_data(
+        env_type_db, name, server, api_key, is_default=True)
+    _save_and_exit(env_db, env_data, save_callable)
+
+
+def env_detail(app, params, env_data):
     """Show details on a Juju environment.
 
     From this view it is possible to start the environment, set it as default,
     edit/remove the environment.
 
-    Receives:
+    Receives a params namedtuple-like object including the following fields:
         - env_type_db: the environments meta information;
         - env_db: the environments database;
-        - save_callable: a function called to save a new environment database;
-        - env_data: the environment data.
+        - jenv_db: the jenv files database;
+        - save_callable: a function called to save a new environment database.
+    See quickstart.cli.params.Params.
+    Also receives the current environment data env_data.
     """
-    env_db = copy.deepcopy(env_db)
+    params = params.copy()
     # All the environment views return a tuple (new_env_db, env_data).
     # Set the env_data to None in the case the user quits the application
     # without selecting an environment to use.
-    app.set_return_value_on_exit((env_db, None))
-    index_view = functools.partial(
-        env_index, app, env_type_db, env_db, save_callable)
-    edit_view = functools.partial(
-        env_edit, app, env_type_db, env_db, save_callable, env_data)
+    app.set_return_value_on_exit((params.env_db, None))
+    index_view = functools.partial(env_index, app, params)
+    edit_view = functools.partial(env_edit, app, params, env_data)
 
-    def use(env_data):
-        # Quit the interactive session returning the (possibly modified)
-        # environment database and the environment data corresponding to the
-        # selected environment.
-        raise ui.AppExit((env_db, env_data))
-
-    def set_default(env_data):
-        # Set this environment as the default one, save the env_db and return
-        # to the index view.
-        env_name = env_data['name']
-        env_db['default'] = env_name
-        save_callable(env_db)
-        app.set_message('{} successfully set as default'.format(env_name))
-        index_view()
-
-    def remove(env_data):
-        # The environment deletion is confirmed: remove the environment from
-        # the database, save the new env_db and return to the index view.
-        env_name = env_data['name']
-        envs.remove_env(env_db, env_name)
-        save_callable(env_db)
-        app.set_message('{} successfully removed'.format(env_name))
-        index_view()
-
-    def confirm_removal(env_data):
-        # Ask confirmation before removing an environment.
-        ui.show_dialog(
-            app, 'Remove the {} environment'.format(env_data['name']),
-            'This action cannot be undone!',
-            actions=[
-                (('control alert', 'confirm'), ui.thunk(remove, env_data)),
-            ],
-        )
-
-    env_metadata = envs.get_env_metadata(env_type_db, env_data)
+    env_metadata = envs.get_env_metadata(params.env_type_db, env_data)
     app.set_title(envs.get_env_short_description(env_data))
     # Validate the environment.
     errors = envs.validate(env_metadata, env_data)
@@ -360,49 +424,180 @@ def env_detail(app, env_type_db, env_db, save_callable, env_data):
         ])
     else:
         # Without errors, it is possible to use/start this environment.
-        controls.append(ui.MenuButton('use', ui.thunk(use, env_data)))
+        use_callback = ui.thunk(_use, params.env_db, env_data)
+        controls.append(ui.MenuButton('use', use_callback))
     app.set_status(status)
     if not env_data['is-default']:
-        controls.append(
-            ui.MenuButton('set default', ui.thunk(set_default, env_data)))
+        # If the environment is not the default one, it is possible to set it
+        # as default.
+        set_default_callback = ui.thunk(
+            _set_default, params.env_db, env_data, params.save_callable,
+            app.set_message, index_view)
+        controls.append(ui.MenuButton('set default', set_default_callback))
+    # Add the controls to edit and remove the environment.
+    remove_callback = ui.thunk(
+        _remove, params.env_db, env_data, params.save_callable,
+        app.set_message, index_view)
+    confirm_removal_callback = ui.thunk(
+        _confirm_removal, app, env_data, remove_callback)
     controls.extend([
         ui.MenuButton('edit', ui.thunk(edit_view)),
-        ui.MenuButton(
-            ('control alert', 'remove'), ui.thunk(confirm_removal, env_data)),
+        ui.MenuButton(('control alert', 'remove'), confirm_removal_callback),
     ])
     widgets.append(ui.create_controls(*controls))
     listbox = urwid.ListBox(urwid.SimpleFocusListWalker(widgets))
     app.set_contents(listbox)
 
 
-def env_edit(app, env_type_db, env_db, save_callable, env_data):
+def _use(env_db, env_data):
+    """Use the selected environment.
+
+    Quit the interactive session returning the (possibly modified)
+    environment database and the environment data corresponding to the
+    selected environment.
+    """
+    raise ui.AppExit((env_db, env_data))
+
+
+def _set_default(env_db, env_data, save_callable, set_message, redirect_view):
+    """Set this environment as the default one.
+
+    Save the env_db and return to the given view.
+    Also output a notification using the given set_message callable.
+    """
+    env_name = env_data['name']
+    env_db['default'] = env_name
+    save_callable(env_db)
+    set_message('{} successfully set as default'.format(env_name))
+    redirect_view()
+
+
+def _confirm_removal(app, env_data, remove_callable):
+    """Ask confirmation before removing an environment."""
+    ui.show_dialog(
+        app, 'Remove the {} environment'.format(env_data['name']),
+        'This action cannot be undone!',
+        actions=[(('control alert', 'confirm'), remove_callable)],
+    )
+
+
+def _remove(env_db, env_data, save_callable, set_message, redirect_view):
+    """Remove the environment represented by env_data from the database.
+
+    Save the new env_db and return to the given view.
+    Also output a notification using the given set_message callable.
+    """
+    env_name = env_data['name']
+    envs.remove_env(env_db, env_name)
+    save_callable(env_db)
+    set_message('{} successfully removed'.format(env_name))
+    redirect_view()
+
+
+def jenv_detail(app, params, env_data):
+    """Show details on a Juju imported environment.
+
+    The environment is not included in the environments.yaml file, but just
+    found in the jenv database.
+    From this view it is possible to start the environment.
+
+    Receives a params namedtuple-like object including the following fields:
+        - env_type_db: the environments meta information;
+        - env_db: the environments database;
+        - jenv_db: the jenv files database;
+        - save_callable: a function called to save a new environment database;
+    See quickstart.cli.params.Params.
+    Also receives the current environment data env_data.
+    """
+    params = params.copy()
+    # All the environment views return a tuple (new_env_db, env_data).
+    # Set the env_data to None in the case the user quits the application
+    # without selecting an environment to use.
+    app.set_return_value_on_exit((params.env_db, None))
+    index_view = functools.partial(env_index, app, params)
+
+    app.set_title(jenv.get_env_short_description(env_data))
+    widgets = []
+    for key, value in jenv.get_env_details(env_data):
+        widgets.append(urwid.Text(['{}: '.format(key), ('highlight', value)]))
+    widgets.extend([
+        urwid.Divider(),
+        urwid.Text([
+            ('highlight', 'Imported active environment.\n'),
+            'This environment is not included in your environments.yaml file.'
+            '\nFor this reason, it is not possible to edit it.\n'
+            'However, you can use the link below to ',
+            ('highlight', 'use Juju Quickstart'),
+            ' with this environment or ',
+            ('highlight', 'remove the corresponding jenv file'),
+            '.\n'
+            'Note that removing the Juju generated environment file does not '
+            'destroy the corresponding active environment.'
+        ]),
+    ])
+
+    remove_callback = ui.thunk(
+        _remove_jenv, params.jenv_db, env_data, params.remove_jenv_callable,
+        app.set_message, index_view)
+    confirm_removal_callback = ui.thunk(
+        _confirm_removal, app, env_data, remove_callback)
+    controls = [
+        ui.MenuButton('back', ui.thunk(index_view)),
+        ui.MenuButton('use', ui.thunk(_use, params.env_db, env_data)),
+        ui.MenuButton(('control alert', 'remove'), confirm_removal_callback),
+    ]
+    widgets.append(ui.create_controls(*controls))
+    listbox = urwid.ListBox(urwid.SimpleFocusListWalker(widgets))
+    app.set_contents(listbox)
+    app.set_status([' \N{RIGHTWARDS ARROW OVER LEFTWARDS ARROW} navigate '])
+
+
+def _remove_jenv(
+        jenv_db, env_data, remove_jenv_callable, set_message, redirect_view):
+    """Remove the jenv file corresonding to the env_data environment.
+
+    Update the provided jenv_db and return to the given view.
+    Also output a notification using the given set_message callable.
+    """
+    env_name = env_data['name']
+    # Remove the jenv file.
+    msg = remove_jenv_callable(env_name)
+    if msg is None:
+        msg = '{} successfully removed'.format(env_name)
+        # Also remove the environments from the jenv database.
+        del jenv_db['environments'][env_name]
+    set_message(msg)
+    redirect_view()
+
+
+def env_edit(app, params, env_data):
     """Create or modify a Juju environment.
 
     This view displays an edit form allowing for environment
     creation/modification. Saving the form redirects to the environment detail
     view if the values are valid.
 
-    Receives:
+    Receives a params namedtuple-like object including the following fields:
         - env_type_db: the environments meta information;
         - env_db: the environments database;
+        - jenv_db: the jenv files database;
         - save_callable: a function called to save a new environment database;
-        - env_data: the environment data.
+    See quickstart.cli.params.Params.
+    Also receives the current environment data env_data.
 
     The last value (env_data) indicates whether this view is used to create a
     new environment or to change an existing one. In the former case, env_data
     does not include the "name" key. If instead the environment already exists,
     env_data includes the "name" key and all the other environment info.
     """
-    env_db = copy.deepcopy(env_db)
+    params = params.copy()
     # All the environment views return a tuple (new_env_db, env_data).
     # Set the env_data to None in the case the user quits the application
     # without selecting an environment to use.
-    app.set_return_value_on_exit((env_db, None))
-    env_metadata = envs.get_env_metadata(env_type_db, env_data)
-    index_view = functools.partial(
-        env_index, app, env_type_db, env_db, save_callable)
-    detail_view = functools.partial(
-        env_detail, app, env_type_db, env_db, save_callable)
+    app.set_return_value_on_exit((params.env_db, None))
+    env_metadata = envs.get_env_metadata(params.env_type_db, env_data)
+    index_view = functools.partial(env_index, app, params)
+    detail_view = functools.partial(env_detail, app, params)
     if 'name' in env_data:
         exists = True
         title = 'Edit the {} environment'
@@ -425,38 +620,6 @@ def env_edit(app, env_type_db, env_db, save_callable, env_data):
         ' field errors ',
     ])
 
-    def save(env_data, get_new_env_data):
-        # Create a new environment or save changes for an existing one.
-        # The new values are saved only if the new env_data is valid, in which
-        # case also redirect to the environment detail view.
-        new_env_data = get_new_env_data()
-        # Validate the new env_data.
-        errors = envs.validate(env_metadata, new_env_data)
-        new_name = new_env_data['name']
-        initial_name = env_data.get('name')
-        if (new_name != initial_name) and new_name in env_db['environments']:
-            errors['name'] = 'an environment with this name already exists'
-        # If errors are found, re-render the form passing the errors. This way
-        # the errors are displayed as part of the form and the user is given
-        # the opportunity to fix the invalid values.
-        if errors:
-            return render_form(new_env_data, errors)
-        # Without errors, normalize the new values, update the env_db and save
-        # the resulting environments database.
-        env_data = envs.normalize(env_metadata, new_env_data)
-        # If this is the only environment in the db, set it as the default one.
-        if not env_db['environments']:
-            env_data['is-default'] = True
-        envs.set_env_data(env_db, initial_name, env_data)
-        save_callable(env_db)
-        verb = 'modified' if exists else 'created'
-        app.set_message('{} successfully {}'.format(new_name, verb))
-        return detail_view(env_data)
-
-    def cancel(env_data):
-        # Dismiss any changes and return to the index or detail view.
-        return detail_view(env_data) if exists else index_view()
-
     def render_form(data, errors):
         # Render the environment edit form.
         widgets = [
@@ -470,9 +633,15 @@ def env_edit(app, env_type_db, env_db, save_callable, env_data):
         form_widgets, get_new_env_data = forms.create_form(
             field_value_pairs, errors)
         widgets.extend(form_widgets)
+        save_callback = ui.thunk(
+            _save, params.env_db, env_data, env_metadata, params.save_callable,
+            exists, get_new_env_data, render_form, app.set_message,
+            detail_view)
+        cancel_callback = ui.thunk(
+            _cancel, env_data, exists, index_view, detail_view)
         actions = (
-            ('save', ui.thunk(save, env_data, get_new_env_data)),
-            ('cancel', ui.thunk(cancel, env_data)),
+            ('save', save_callback),
+            ('cancel', cancel_callback),
             ('restore', ui.thunk(render_form, env_data, initial_errors)),
         )
         widgets.append(forms.create_actions(actions))
@@ -482,3 +651,41 @@ def env_edit(app, env_type_db, env_db, save_callable, env_data):
 
     # Render the initial form.
     render_form(env_data, initial_errors)
+
+
+def _save(
+        env_db, env_data, env_metadata, save_callable, exists,
+        get_new_env_data, render_form, set_message, redirect_view):
+    """Create a new environment or save changes for an existing one.
+
+    The new values are saved only if the new env_data is valid, in which
+    case also redirect to the given view.
+    """
+    new_env_data = get_new_env_data()
+    # Validate the new env_data.
+    errors = envs.validate(env_metadata, new_env_data)
+    new_name = new_env_data['name']
+    initial_name = env_data.get('name')
+    if (new_name != initial_name) and new_name in env_db['environments']:
+        errors['name'] = 'an environment with this name already exists'
+    # If errors are found, re-render the form passing the errors. This way
+    # the errors are displayed as part of the form and the user is given
+    # the opportunity to fix the invalid values.
+    if errors:
+        return render_form(new_env_data, errors)
+    # Without errors, normalize the new values, update the env_db and save
+    # the resulting environments database.
+    env_data = envs.normalize(env_metadata, new_env_data)
+    # If this is the only environment in the db, set it as the default one.
+    if not env_db['environments']:
+        env_data['is-default'] = True
+    envs.set_env_data(env_db, initial_name, env_data)
+    save_callable(env_db)
+    verb = 'modified' if exists else 'created'
+    set_message('{} successfully {}'.format(new_name, verb))
+    return redirect_view(env_data)
+
+
+def _cancel(env_data, exists, index_view, detail_view):
+    """Dismiss any changes and return to the index or detail view."""
+    return detail_view(env_data) if exists else index_view()
