@@ -19,16 +19,25 @@
 from __future__ import unicode_literals
 
 import logging
+import ssl
 
 import jujuclient
-import ssl
 import websocket
 
 
+# Define the constant tag used by Juju for user entities.
+JUJU_USER_TAG = 'user'
+OPCODE_TEXT = websocket.ABNF.OPCODE_TEXT
+# Define the options used to initiate the secure WebSocket connection.
+SSLOPT = {
+    'cert_reqs': ssl.CERT_NONE,
+    'ssl_version': ssl.PROTOCOL_TLSv1,
+}
+
+
 def connect(api_url):
     """Return an Environment instance connected to the given API URL."""
-    connection = WebSocketConnection(
-        sslopt={'ssl_version': ssl.PROTOCOL_TLSv1})
+    connection = WebSocketConnection(sslopt=SSLOPT)
     # See the websocket.create_connection function.
     connection.settimeout(websocket.default_timeout)
     connection.connect(api_url, origin=api_url)
@@ -43,6 +52,15 @@ class Environment(jujuclient.Environment):
     deployments to specific machines.
     """
 
+    def authenticate(self, username, password):
+        """Log in into the Juju environment with the given credentials.
+
+        The user name must be provided without the default Juju prefix for
+        users. Use self.login to log in specifying the complete user tag.
+        """
+        return self.login(
+            password, user='{}-{}'.format(JUJU_USER_TAG, username))
+
     def deploy_bundle(self, yaml, name=None, bundle_id=None):
         """Deploy a bundle."""
         params = {'YAML': yaml}
@@ -124,19 +142,23 @@ class Environment(jujuclient.Environment):
 class WebSocketConnection(websocket.WebSocket):
     """A WebSocket client connection."""
 
-    def send(self, message):
+    def send(self, payload, opcode=OPCODE_TEXT):
         """Send the given WebSocket message.
 
-        Overridden to add logging.
+        Overridden to add logging in the case the payload is text.
         """
-        logging.debug('API message: --> {}'.format(message.decode('utf-8')))
-        return super(WebSocketConnection, self).send(message)
+        if opcode == OPCODE_TEXT:
+            message = payload.decode('utf-8')
+            logging.debug('API message: --> {}'.format(message))
+        return super(WebSocketConnection, self).send(payload, opcode=opcode)
 
     def recv(self):
         """Receive a message from the WebSocket server.
 
-        Overridden to add logging.
+        Overridden to add logging in the case the received data is text.
         """
-        message = super(WebSocketConnection, self).recv()
-        logging.debug('API message: <-- {}'.format(message.decode('utf-8')))
-        return message
+        data = super(WebSocketConnection, self).recv()
+        if isinstance(data, bytes):
+            message = data.decode('utf-8')
+            logging.debug('API message: <-- {}'.format(message))
+        return data
