@@ -32,15 +32,20 @@ import webbrowser
 import quickstart
 from quickstart import (
     app,
+    netutils,
     packaging,
     platform_support,
     settings,
     utils,
 )
-from quickstart.cli import views
+from quickstart.cli import (
+    params,
+    views,
+)
 from quickstart.models import (
     charms,
     envs,
+    jenv,
 )
 
 
@@ -110,7 +115,7 @@ def _validate_bundle(options, parser):
     if bundle.startswith('http://') or bundle.startswith('https://'):
         # Load the bundle from a remote URL.
         try:
-            bundle_yaml = utils.urlread(bundle)
+            bundle_yaml = netutils.urlread(bundle)
         except IOError as err:
             return parser.error('unable to open bundle URL: {}'.format(err))
     else:
@@ -201,16 +206,21 @@ def _create_save_callable(parser, env_file):
     return save_callable
 
 
-def _start_interactive_session(parser, env_type_db, env_db, env_file):
+def _start_interactive_session(parser, env_type_db, env_db, jenv_db, env_file):
     """Start the Urwid interactive session.
 
     Return the env_data corresponding to the user selected environment.
     Exit the application if the user exits the interactive session without
     selecting an environment to start.
     """
-    save_callable = _create_save_callable(parser, env_file)
-    new_env_db, env_data = views.show(
-        views.env_index, env_type_db, env_db, save_callable)
+    parameters = params.Params(
+        env_type_db=env_type_db,
+        env_db=env_db,
+        jenv_db=jenv_db,
+        save_callable=_create_save_callable(parser, env_file),
+        remove_jenv_callable=jenv.remove,
+    )
+    new_env_db, env_data = views.show(views.env_index, parameters)
     if new_env_db != env_db:
         print('changes to the environments file have been saved')
     if env_data is None:
@@ -221,16 +231,23 @@ def _start_interactive_session(parser, env_type_db, env_db, env_file):
     return env_data
 
 
-def _retrieve_env_data(parser, env_type_db, env_db, env_name):
+def _retrieve_env_data(parser, env_type_db, env_db, jenv_db, env_name):
     """Retrieve and return the env_data corresponding to the given env_name.
 
     Invoke a parser error if the environment does not exist or is not valid.
     """
     try:
         env_data = envs.get_env_data(env_db, env_name)
-    except ValueError as err:
-        # The specified environment does not exist.
-        return parser.error(bytes(err))
+    except ValueError:
+        # The specified environment does not exist in the environments file.
+        # Check if this is an imported environment.
+        try:
+            return envs.get_env_data(jenv_db, env_name)
+        except ValueError as err:
+            # The environment cannot be found anywhere, exit with an error.
+            return parser.error(bytes(err))
+    # If the environment was found in the environments.yaml file, we can also
+    # validate it.
     env_metadata = envs.get_env_metadata(env_type_db, env_data)
     errors = envs.validate(env_metadata, env_data)
     if errors:
@@ -274,17 +291,19 @@ def _setup_env(options, parser):
 
     # Retrieve the environment database (or create an in-memory empty one).
     env_db = _retrieve_env_db(parser, env_file if env_file_exists else None)
+    jenv_db = jenv.get_env_db()
 
     # Validate the environment.
     env_type_db = envs.get_env_type_db()
     if interactive:
         # Start the interactive session.
         env_data = _start_interactive_session(
-            parser, env_type_db, env_db, env_file)
+            parser, env_type_db, env_db, jenv_db, env_file)
     else:
         # This is a non-interactive session and we need to validate the
         # selected environment before proceeding.
-        env_data = _retrieve_env_data(parser, env_type_db, env_db, env_name)
+        env_data = _retrieve_env_data(
+            parser, env_type_db, env_db, jenv_db, env_name)
     # Check for local support, if requested.
     options.env_type = env_data['type']
     no_local_support = not platform_support.supports_local(options.platform)
@@ -293,7 +312,6 @@ def _setup_env(options, parser):
             'this host platform does not support local environments'
         )
     # Update the options namespace with the new values.
-    options.admin_secret = env_data.get('admin-secret')
     options.env_file = env_file
     options.env_name = env_data['name']
     options.default_series = env_data.get('default-series')
@@ -343,10 +361,9 @@ def setup():
     """Set up the application options and logger.
 
     Return the options as a namespace containing the following attributes:
-        - admin_secret: the password to use to access the Juju API or None if
-          no admin-secret is present in the $JUJU_HOME/environment.yaml file;
         - bundle: the optional bundle (path or URL) to be deployed;
         - charm_url: the Juju GUI charm URL or None if not specified;
+        - constraints: the environment constrains or None if not set;
         - debug: whether debug mode is activated;
         - distro_only: install Juju only using the distribution packages;
         - env_file: the absolute path of the Juju environments.yaml file;
@@ -357,8 +374,7 @@ def setup():
         - platform: The host platform;
         - upload_tools: whether to upload local version of tools;
         - upload_series: the comma-separated series list for which tools will
-          be uploaded, or None if not set;
-        - constraints: the environment constrains or None if not set.
+          be uploaded, or None if not set.
 
     The following attributes will also be included in the namespace if a bundle
     deployment is requested:
@@ -496,54 +512,68 @@ def run(options):
         print('contents loaded for bundle {} (services: {})'.format(
             options.bundle_name, len(options.bundle_services)))
 
-    juju_command = platform_support.get_juju_command(options.platform)
+    juju_command, custom_juju = platform_support.get_juju_command(
+        options.platform)
 
     logging.debug('ensuring juju and dependencies are installed')
     juju_version = app.ensure_dependencies(
         options.distro_only, options.platform, juju_command)
+    app.check_juju_supported(juju_version)
 
     logging.debug('ensuring SSH keys are available')
     app.ensure_ssh_keys()
 
-    print('bootstrapping the {} environment (type: {})'.format(
-        options.env_name, options.env_type))
-    requires_sudo = False
-    if options.env_type == 'local':
-        # If this is a local environment, notify the user that "sudo" will be
-        # required to bootstrap the application, even in newer Juju versions
-        # where "sudo" is invoked by juju-core itself.
-        print('sudo privileges will be required to bootstrap the environment')
-        # If the Juju version is less than 1.17.2 then use sudo for local envs.
-        requires_sudo = juju_version < (1, 17, 2)
+    # Bootstrap the Juju environment or reuse an already bootstrapped one.
+    already_bootstrapped = True
+    env_type = options.env_type
+    api_url = app.check_bootstrapped(options.env_name)
+    if api_url is None:
+        print('bootstrapping the {} environment'.format(options.env_name))
+        if env_type == 'local':
+            # If this is a local environment, notify the user that "sudo" will
+            # be required by Juju to bootstrap the environment.
+            print('sudo privileges will be required to bootstrap the '
+                  'environment')
+        already_bootstrapped = app.bootstrap(
+            options.env_name, juju_command,
+            debug=options.debug,
+            upload_tools=options.upload_tools,
+            upload_series=options.upload_series,
+            constraints=options.constraints)
+    if already_bootstrapped:
+        print('reusing the already bootstrapped {} environment'.format(
+            options.env_name))
 
-    already_bootstrapped, bootstrap_node_series = app.bootstrap(
-        options.env_name, juju_command,
-        requires_sudo=requires_sudo, debug=options.debug,
-        upload_tools=options.upload_tools,
-        upload_series=options.upload_series,
-        constraints=options.constraints)
+    # Retrieve the environment status, ensure it is in a ready state and
+    # contextually fetch the bootstrap node series.
+    print('retrieving the environment status')
+    bootstrap_node_series = app.status(options.env_name, juju_command)
+
+    # If the environment was not already bootstrapped, we need to retrieve
+    # the API address.
+    if api_url is None:
+        print('retrieving the Juju API address')
+        api_url = app.get_api_url(options.env_name, juju_command)
 
     # Retrieve the admin-secret for the current environment.
-    try:
-        admin_secret = app.get_value_from_jenv(
-            options.env_name, settings.JUJU_HOME, 'admin-secret')
-    except ValueError as err:
-        admin_secret = options.admin_secret
-        if admin_secret is None:
-            # The admin-secret cannot be found in the jenv file and is not
-            # explicitly specified in the environments.yaml file.
-            msg = b'{} or {}'.format(err, options.env_file.encode('utf-8'))
-            raise app.ProgramExit(msg)
+    print('retrieving the Juju environment credentials')
+    username, password = app.get_credentials(options.env_name)
 
-    print('retrieving the Juju API address')
-    api_url = app.get_api_url(options.env_name, juju_command)
     print('connecting to {}'.format(api_url))
-    env = app.connect(api_url, admin_secret)
+    env = app.connect(api_url, username, password)
+
+    if already_bootstrapped:
+        # Retrieve the environment type from the live environment: it may be
+        # different from the one declared on the environments.yaml file.
+        # Moreover, an imported jenv file could be in use, in which case the
+        # environment type is probably still unknown.
+        env_type = app.get_env_type(env)
+    print('environment type: {}'.format(env_type))
 
     # Inspect the environment and deploy the charm if required.
     charm_url, machine, service_data, unit_data = app.check_environment(
         env, settings.JUJU_GUI_SERVICE_NAME, options.charm_url,
-        options.env_type, bootstrap_node_series, already_bootstrapped)
+        env_type, bootstrap_node_series, already_bootstrapped)
     unit_name = app.deploy_gui(
         env, settings.JUJU_GUI_SERVICE_NAME, charm_url, machine,
         service_data, unit_data)
@@ -552,10 +582,11 @@ def run(options):
     address = app.watch(env, unit_name)
     env.close()
     url = 'https://{}'.format(address)
-    print('\nJuju GUI URL: {}\npassword: {}\n'.format(url, admin_secret))
+    print('\nJuju GUI URL: {}\nusername: {}\npassword: {}\n'.format(
+        url, username, password))
     gui_api_url = 'wss://{}:443/ws'.format(address)
     print('connecting to the Juju GUI server')
-    gui_env = app.connect(gui_api_url, admin_secret)
+    gui_env = app.connect(gui_api_url, username, password)
 
     # Handle bundle deployment.
     if options.bundle is not None:
@@ -584,9 +615,7 @@ def run(options):
         'Run "juju quickstart -i" if you want to manage\n'
         'or bootstrap your Juju environments using the\n'
         'interactive session.\n'
-        'Run "{sudo}juju destroy-environment {eflag}{env_name} [-y]"\n'
-        'to destroy the environment you just bootstrapped.'.format(
-            env_name=options.env_name,
-            sudo='sudo ' if requires_sudo else '',
-            eflag='-e ' if juju_version < (1, 17, 2) else '')
+        'Run "juju destroy-environment {env_name} [-y]"\n'
+        'to destroy the environment you just bootstrapped.'
+        ''.format(env_name=options.env_name)
     )
