@@ -134,23 +134,6 @@ def create_empty_env_db():
     return {'environments': {}}
 
 
-def _load_file(env_file):
-    """Load the given file and return the YAML-parsed contents."""
-    # Load the Juju environments file.
-    try:
-        environments_file = open(env_file.encode('utf-8'))
-    except IOError as err:
-        msg = b'unable to open environments file: {}'.format(err)
-        raise ValueError(msg)
-    # Parse the Juju environments file.
-    try:
-        contents = serializers.yaml_load(environments_file)
-    except Exception as err:
-        msg = b'unable to parse environments file {}: {}'
-        raise ValueError(msg.format(env_file.encode('utf-8'), err))
-    return contents
-
-
 def load(env_file):
     """Load and parse the provided Juju environments.yaml file.
 
@@ -174,7 +157,7 @@ def load(env_file):
         - the environment file contents are not parsable by YAML;
         - the YAML contents are not properly structured.
     """
-    contents = _load_file(env_file)
+    contents = serializers.yaml_load_from_path(env_file)
     if contents is None:
         return create_empty_env_db()
     # Retrieve the environment list.
@@ -199,18 +182,6 @@ def load(env_file):
     return env_db
 
 
-def load_generated(env_file, section='bootstrap-config'):
-    """Given the path to a YAML file, load the file and return the section."""
-    contents = _load_file(env_file)
-    try:
-        section_contents = contents[section]
-    except (KeyError, TypeError):
-        msg = 'invalid YAML contents in {}: {}'.format(env_file, contents)
-        raise ValueError(msg.encode('utf-8'))
-
-    return section_contents
-
-
 def save(env_file, env_db, backup_function=None):
     """Save the given env_db to the provided environments.yaml file.
 
@@ -338,7 +309,7 @@ def set_env_data(env_db, env_name, env_data):
 
 
 def create_local_env_data(env_type_db, name, is_default=False):
-    """Create and return an local (LXC) env_data.
+    """Create and return a local (LXC) env_data.
 
     Local environments' fields (except for name and type) are assumed to be
     either optional or suitable for automatic generation of their values. For
@@ -367,6 +338,29 @@ def create_local_env_data(env_type_db, name, is_default=False):
     return env_data
 
 
+def create_maas_env_data(env_type_db, name, server, api_key, is_default=False):
+    """Create and return a MAAS (bare metal) env_data.
+
+    The env_data is created using the given name, MAAS server and MAAS OAuth
+    API key.
+    """
+    env_data = {
+        'type': 'maas',
+        'name': name,
+        'is-default': is_default,
+        'maas-server': server,
+        'maas-oauth': api_key,
+    }
+    env_metadata = get_env_metadata(env_type_db, env_data)
+    # Retrieve a list of missing required fields.
+    missing_fields = [
+        field for field in env_metadata['fields']
+        if field.required and field.name not in env_data]
+    # Assume all missing fields can be automatically generated.
+    env_data.update((field.name, field.generate()) for field in missing_fields)
+    return env_data
+
+
 def remove_env(env_db, env_name):
     """Remove the environment named env_name from the environments database.
 
@@ -462,9 +456,8 @@ def get_env_type_db():
         ),
         'fields': (
             provider_field,
+            # Required fields.
             name_field,
-            admin_secret_field,
-            default_series_field,
             fields.PasswordField(
                 'access-key', label='access key', required=True,
                 help='The access key to use to authenticate to AWS. '
@@ -481,12 +474,15 @@ def get_env_type_db():
                 'secret-key', label='secret key', required=True,
                 help='The secret key to use to authenticate to AWS. '
                      'See the access key help above.'),
-            fields.AutoGeneratedStringField(
-                'control-bucket', label='control bucket', required=False,
-                help='the globally unique S3 bucket name'),
+            # Optional fields.
             fields.ChoiceField(
                 'region', choices=ec2_regions, default='us-east-1',
                 label='region', required=False, help='the ec2 region to use'),
+            admin_secret_field,
+            default_series_field,
+            fields.AutoGeneratedStringField(
+                'control-bucket', label='control bucket', required=False,
+                help='the globally unique S3 bucket name'),
             is_default_field,
         ),
     }
@@ -502,20 +498,8 @@ def get_env_type_db():
         ),
         'fields': (
             provider_field,
+            # Required fields.
             name_field,
-            admin_secret_field,
-            default_series_field,
-            fields.BoolField(
-                'use-floating-ip', label='use floating IP', allow_mixed=False,
-                required=True,
-                help='Specifies whether the use of a floating IP address is '
-                     'required to give the nodes a public IP address. '
-                     'Some installations assign public IP addresses by '
-                     'default without requiring a floating IP address. '
-                     'For HP Public Cloud, floating IP must be activated.'),
-            fields.AutoGeneratedStringField(
-                'control-bucket', label='control bucket', required=False,
-                help='the globally unique swift bucket name'),
             fields.SuggestionsStringField(
                 'auth-url', label='authentication URL', required=True,
                 help='The Keystone URL to use in the authentication process. '
@@ -527,11 +511,20 @@ def get_env_type_db():
                 help='The OpenStack tenant name. For HP Public Cloud, this is '
                      'listed as the project name on the '
                      'https://horizon.hpcloud.com/landing/ page.'),
+            fields.BoolField(
+                'use-floating-ip', label='use floating IP', allow_mixed=False,
+                required=True,
+                help='Specifies whether the use of a floating IP address is '
+                     'required to give the nodes a public IP address. '
+                     'Some installations assign public IP addresses by '
+                     'default without requiring a floating IP address. '
+                     'For HP Public Cloud, floating IP must be activated.'),
             fields.SuggestionsStringField(
                 'region', label='region', required=True,
                 help='The OpenStack region to use. '
                      'For HP Public Cloud, use one of the following:',
                 suggestions=hp_regions),
+            # Optional fields.
             fields.ChoiceField(
                 'auth-mode', label='authentication mode', required=False,
                 default='userpass', choices=('userpass', 'keypair'),
@@ -555,6 +548,11 @@ def get_env_type_db():
             fields.PasswordField(
                 'secret-key', label='secret key', required=False,
                 help='the secret key to use for the keypair authentication'),
+            fields.AutoGeneratedStringField(
+                'control-bucket', label='control bucket', required=False,
+                help='the globally unique swift bucket name'),
+            admin_secret_field,
+            default_series_field,
             is_default_field,
         ),
     }
@@ -570,12 +568,8 @@ def get_env_type_db():
         ),
         'fields': (
             provider_field,
+            # Required fields.
             name_field,
-            admin_secret_field,
-            default_series_field,
-            fields.ChoiceField(
-                'location', choices=azure_locations, label='location',
-                required=True, help='the region to use'),
             fields.StringField(
                 'management-subscription-id', required=True,
                 label='management subscription ID',
@@ -596,6 +590,12 @@ def get_env_type_db():
                      'https://manage.windowsazure.com (Storage). '
                      'You must create the storage account in the same '
                      'region/location specified by the location key value.'),
+            fields.ChoiceField(
+                'location', choices=azure_locations, label='location',
+                required=True, help='the region to use'),
+            # Optional fields.
+            admin_secret_field,
+            default_series_field,
             is_default_field,
         ),
     }
@@ -612,8 +612,7 @@ def get_env_type_db():
         'fields': (
             provider_field,
             name_field,
-            admin_secret_field,
-            default_series_field,
+            # SDC fields.
             fields.StringField(
                 'sdc-user', required=True,
                 label='SDC user name',
@@ -628,6 +627,7 @@ def get_env_type_db():
                 'sdc-url', choices=joyent_sdc_urls, label='SDC URL',
                 required=False, default=joyent_sdc_urls[0],
                 help='the region URL to use for SDC'),
+            # Manta fields.
             fields.StringField(
                 'manta-user', required=True,
                 label='manta user name',
@@ -643,6 +643,7 @@ def get_env_type_db():
                 'manta-url', label='manta URL', required=False,
                 default='https://us-east.manta.joyent.com',
                 help='the region URL to use for manta'),
+            # Optional fields.
             fields.FilePathField(
                 'private-key-path', label='private key path', required=False,
                 default='~/.ssh/id_rsa',
@@ -650,6 +651,90 @@ def get_env_type_db():
             fields.StringField(
                 'algorithm', label='SSH algorithm', required=False,
                 default='rsa-sha256'),
+            admin_secret_field,
+            default_series_field,
+            is_default_field,
+        ),
+    }
+    env_type_db['maas'] = {
+        'label': 'MAAS (bare metal)',
+        'description': (
+            'Metal As A Service is software which allows you to deal with '
+            'physical hardware just as easily as virtual nodes. MAAS lets you '
+            'treat physical servers like virtual machines in the cloud. '
+            'Rather than having to manage each server individually, MAAS '
+            'turns your bare metal into an elastic cloud-like resource. '
+            'Specifically, MAAS allows for services to be deployed to bare '
+            'metal via Juju. '
+            'See https://juju.ubuntu.com/docs/config-maas.html and '
+            'http://maas.ubuntu.com/ for more information about MAAS.'
+        ),
+        'fields': (
+            provider_field,
+            # Required fields.
+            name_field,
+            fields.StringField(
+                'maas-server', label='server address', required=True,
+                help='The MAAS server address, for instance '
+                     'http://<my-maas-server>/MAAS.'),
+            fields.PasswordField(
+                'maas-oauth', label='API key', required=True,
+                help='The MAAS API key. This can be found by going to the '
+                     'MAAS user preferences page (/MAAS/account/prefs/). A '
+                     'link to the account preferences is accessible from the '
+                     'drop-down menu that appears when clicking your user '
+                     'name at the top-right of the page.'),
+            # Optional fields.
+            fields.FilePathField(
+                'authorized-keys-path', required=False,
+                label='SSH public key file',
+                help='If keys are specified, MAAS will be able to '
+                     'automatically add them to each unit it manages.'),
+            admin_secret_field,
+            default_series_field,
+            is_default_field,
+        ),
+    }
+    env_type_db['manual'] = {
+        'label': 'Manual Provisioning',
+        'description': (
+            'Juju provides a feature called "Manual Provisioning" that '
+            'enables you to deploy Juju, and charms, to existing systems. '
+            'This is useful if you have groups of machines that you want to '
+            "use for Juju but don't want to add the complexity of a new "
+            'OpenStack or MAAS setup. It is also useful as a means of '
+            'deploying workloads to VPS providers and other cheap hosting '
+            'options. Manual provisioning enables you to run Juju on systems '
+            'that have a supported operating system installed. You will need '
+            'to ensure that you have both SSH access and sudo rights. '
+            'See https://juju.ubuntu.com/docs/config-manual.html for more '
+            'details on the manual provider configuration, including its '
+            'caveats and limitattions.'
+        ),
+        'fields': (
+            provider_field,
+            # Required fields.
+            name_field,
+            fields.StringField(
+                'bootstrap-host', label='bootstrap host', required=True,
+                help='The host name/IP address of the machine where the '
+                     'bootstrap machine agent will be started.'),
+            # Optional fields.
+            fields.StringField(
+                'bootstrap-user', label='bootstrap user', required=False,
+                help='The user name to authenticate as when connecting to the '
+                     'bootstrap machine. It defaults to the current user.'),
+            fields.StringField(
+                'storage-listen-ip', label='storage listen IP', required=False,
+                help="The IP address that the bootstrap machine's Juju "
+                     'storage server will listen on. By default, storage will '
+                     'be served on all network interfaces.'),
+            fields.IntField(
+                'storage-port', min_value=1, max_value=65535,
+                label='storage port', required=False, default=8040,
+                help="The TCP port that the bootstrap machine's Juju storage "
+                     'server will listen on.'),
+            admin_secret_field,
             is_default_field,
         ),
     }
@@ -664,8 +749,18 @@ def get_env_type_db():
         'fields': (
             provider_field,
             name_field,
-            admin_secret_field,
-            default_series_field,
+            # LXC related fields.
+            fields.BoolField(
+                'lxc-clone', label='use lxc-clone', required=False,
+                help='Use lxc-clone to create the containers used as '
+                     'machines. LXC clone is enabled by default for Trusty '
+                     'and above, and disabled for earlier Ubuntu releases.'),
+            fields.BoolField(
+                'lxc-clone-aufs', label='use lxc-clone-aufs', required=False,
+                help='Use aufs as a backing-store for the LXC clones. Note '
+                     "that there are some situations where aufs doesn't "
+                     'entirely behave as intuitively as one might expect. '
+                     'If not specified, aufs is not used.'),
             fields.StringField(
                 'root-dir', label='root dir', required=False,
                 default='~/.juju/local',
@@ -683,17 +778,9 @@ def get_env_type_db():
             fields.StringField(
                 'network-bridge', label='network bridge', required=False,
                 default='lxcbr0', help='the LXC bridge interface to use'),
-            fields.BoolField(
-                'lxc-clone', label='use lxc-clone', required=False,
-                help='Use lxc-clone to create the containers used as '
-                     'machines. LXC clone is enabled by default for Trusty '
-                     'and above, and disabled for earlier Ubuntu releases.'),
-            fields.BoolField(
-                'lxc-clone-aufs', label='use lxc-clone-aufs', required=False,
-                help='Use aufs as a backing-store for the LXC clones. Note '
-                     "that there are some situations where aufs doesn't "
-                     'entirely behave as intuitively as one might expect. '
-                     'If not specified, aufs is not used.'),
+            # Other optional fields.
+            admin_secret_field,
+            default_series_field,
             is_default_field,
         ),
     }
