@@ -29,17 +29,19 @@ OSX = object()
 UNKNOWN_PLATFORM = object()
 WINDOWS = object()
 
-# The base charmworld API URL containing information about charms.
+# The base charm store API URL containing information about charms.
 # This URL must be formatted with a series and a charm name.
-CHARMWORLD_API = 'http://manage.jujucharms.com/api/3/charm/{series}/{charm}'
+CHARMSTORE_API = (
+    'https://api.jujucharms.com'
+    '/charmstore/v4/{series}/{charm}/meta/id')
 
 # The default Juju GUI charm URLs for each supported series. Used when it is
-# not possible to retrieve the charm URL from the charmworld API, e.g. due to
-# temporary connection/charmworld errors.
+# not possible to retrieve the charm URL from the charm store API, e.g. due to
+# temporary connection/charm store errors.
 # Keep this list sorted by release date (older first).
 DEFAULT_CHARM_URLS = collections.OrderedDict((
-    ('precise', 'cs:precise/juju-gui-90'),
-    ('trusty', 'cs:trusty/juju-gui-2'),
+    ('precise', 'cs:precise/juju-gui-104'),
+    ('trusty', 'cs:trusty/juju-gui-16'),
 ))
 
 # The quickstart app short description.
@@ -61,7 +63,8 @@ JUJU_PACKAGES = {
 }
 
 # The possible values for the environments.yaml default-series field.
-JUJU_DEFAULT_SERIES = ('precise', 'quantal', 'raring', 'saucy', 'trusty')
+JUJU_DEFAULT_SERIES = (
+    'precise', 'quantal', 'raring', 'saucy', 'trusty', 'utopic', 'vivid')
 
 # Retrieve the current juju-core home.
 JUJU_HOME = os.getenv('JUJU_HOME', '~/.juju')
@@ -75,6 +78,12 @@ JUJU_GUI_SERVICE_NAME = JUJU_GUI_CHARM_NAME
 # The set of series supported by the Juju GUI charm.
 JUJU_GUI_SUPPORTED_SERIES = tuple(DEFAULT_CHARM_URLS.keys())
 
+# The minimum Juju version supported by Juju Quickstart,
+JUJU_SUPPORTED_VERSION = (1, 18, 1)
+
+# The path to the MAAS command line interface.
+MAAS_CMD = '/usr/bin/maas'
+
 # The minimum Juju GUI charm revision supporting bundle deployments, for each
 # supported series. Assume not listed series to always support bundles.
 MINIMUM_REVISIONS_FOR_BUNDLES = collections.defaultdict(
