@@ -18,7 +18,10 @@
 
 from __future__ import unicode_literals
 
-from contextlib import contextmanager
+from contextlib import (
+    contextmanager,
+    nested,
+)
 import unittest
 
 import mock
@@ -28,14 +31,23 @@ from quickstart import settings
 from quickstart.cli import (
     base,
     forms,
+    params,
     ui,
     views,
 )
-from quickstart.models import envs
+from quickstart.models import (
+    envs,
+    jenv,
+)
 from quickstart.tests import helpers
 from quickstart.tests.cli import helpers as cli_helpers
 
 
+MAAS_NAME = 'maas-name'
+MAAS_SERVER = 'http://1.2.3.4/MAAS'
+MAAS_API_KEY = 'maas-secret'
+
+
 def local_envs_supported(value):
     """Simulate local environments support in the current platform.
 
@@ -48,6 +60,26 @@ def local_envs_supported(value):
         mock.Mock(return_value=value))
 
 
+def maas_env_detected(value):
+    """Simulate whether a logged in MAAS remote API has been detected.
+
+    The value argument is a boolean representing whether or not a MAAS API
+    is available. If available, also patch "maas list" to return the following
+    default values: MAAS_NAME, MAAS_SERVER and MAAS_API_KEY.
+    Return a context manager that can be used when calling views.
+    """
+    patch_cli_available = mock.patch(
+        'quickstart.cli.views.maas.cli_available',
+        mock.Mock(return_value=value))
+    if value:
+        return_value = (MAAS_NAME, MAAS_SERVER, MAAS_API_KEY)
+        patch_get_api_info = mock.patch(
+            'quickstart.cli.views.maas.get_api_info',
+            mock.Mock(return_value=return_value))
+        return nested(patch_cli_available, patch_get_api_info)
+    return patch_cli_available
+
+
 class TestShow(unittest.TestCase):
 
     @contextmanager
@@ -82,8 +114,10 @@ class TestShow(unittest.TestCase):
         view = mock.Mock()
         run_side_effect = ui.AppExit('bad wolf')
         with self.patch_setup_urwid_app(run_side_effect=run_side_effect):
-            return_value = views.show(view)
+            with helpers.mock_print as mock_print:
+                return_value = views.show(view)
         self.assertEqual('bad wolf', return_value)
+        mock_print.assert_called_once_with('interactive session closed')
 
     def test_view_arguments(self):
         # The view is invoked passing the app and all the optional show
@@ -103,6 +137,7 @@ class EnvViewTestsMixin(cli_helpers.CliAppTestsMixin):
         # Set up the base Urwid application.
         self.loop, self.app = base.setup_urwid_app()
         self.save_callable = mock.Mock()
+        self.remove_jenv_callable = mock.Mock(return_value=None)
 
     def get_widgets_in_contents(self, filter_function=None):
         """Return a list of widgets included in the app contents.
@@ -133,19 +168,70 @@ class EnvViewTestsMixin(cli_helpers.CliAppTestsMixin):
         """
         return lambda arg: isinstance(arg, cls)
 
+    def make_params(self, env_db, jenv_db):
+        """Create and return view parameters using the given env databases."""
+        return params.Params(
+            env_type_db=self.env_type_db,
+            env_db=env_db,
+            jenv_db=jenv_db,
+            save_callable=self.save_callable,
+            remove_jenv_callable=self.remove_jenv_callable,
+        )
+
+    def click_remove_button(self, env_name):
+        """Click the remove button in an environment detail view.
+
+        Assume the view was already called.
+        Return the dialog button widgets and the original view contents.
+        """
+        original_contents = self.app.get_contents()
+        # The "remove" button is the last one.
+        remove_button = self.get_control_buttons()[-1]
+        cli_helpers.emit(remove_button)
+        # The original env detail contents have been replaced.
+        contents = self.app.get_contents()
+        self.assertIsNot(contents, original_contents)
+        # A "remove" confirmation dialog is displayed.
+        title_widget, message_widget, buttons = cli_helpers.inspect_dialog(
+            contents)
+        self.assertEqual(
+            'Remove the {} environment'.format(env_name), title_widget.text)
+        self.assertEqual('This action cannot be undone!', message_widget.text)
+        return buttons, original_contents
+
+    def cancel_removal(self, env_name):
+        """Cancel the environment deletion by clicking the "cancel" button."""
+        buttons, original_contents = self.click_remove_button(env_name)
+        # The "cancel" button is the first one in the dialog.
+        cancel_button = buttons[0]
+        cli_helpers.emit(cancel_button)
+        return original_contents
+
+    def confirm_removal(self, env_name):
+        """Confirm the environment deletion."""
+        buttons, _ = self.click_remove_button(env_name)
+        # The "confirm" button is the second one in the dialog.
+        confirm_button = buttons[1]
+        cli_helpers.emit(confirm_button)
+
 
 class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
 
     base_status = ' \N{UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW} navigate '
     create_local_caption = (
         '\N{BULLET} automatically create and bootstrap a local environment')
+    create_maas_caption = (
+        '\N{BULLET} automatically create and bootstrap the {} MAAS '
+        'environment'.format(MAAS_NAME))
+    active_environments_message = 'Other active environments'
 
     def test_view_default_return_value_on_exit(self):
         # The view configures the app so that the return value on user exit is
         # a tuple including a copy of the given env_db and None, the latter
         # meaning no environment has been selected.
         env_db = helpers.make_env_db()
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         new_env_db, env_data = self.get_on_exit_return_value(self.loop)
         self.assertEqual(env_db, new_env_db)
         self.assertIsNot(env_db, new_env_db)
@@ -154,7 +240,8 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
     def test_view_title(self):
         # The application title is correctly set up.
         env_db = helpers.make_env_db()
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         self.assertEqual(
             'Select an existing Juju environment or create a new one',
             self.app.get_title())
@@ -162,7 +249,8 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
     def test_view_title_no_environments(self):
         # The application title changes if the env_db has no environments.
         env_db = {'environments': {}}
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         self.assertEqual(
             'No Juju environments already set up: please create one',
             self.app.get_title())
@@ -171,9 +259,9 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
         # The view displays a list of the environments in env_db, and buttons
         # to create new environments.
         env_db = helpers.make_env_db()
+        jenv_db = {'environments': {}}
         with local_envs_supported(True):
-            views.env_index(
-                self.app, self.env_type_db, env_db, self.save_callable)
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
         # A button is created for each existing environment (see details) and
@@ -182,13 +270,49 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
         expected_buttons_number = len(env_db['environments']) + len(env_types)
         self.assertEqual(expected_buttons_number, len(buttons))
 
+    def test_view_contents_with_imported_envs(self):
+        # The view displays a list of active imported environments, and buttons
+        # to create new environments.
+        env_db = {'environments': {}}
+        jenv_db = helpers.make_jenv_db()
+        with local_envs_supported(True):
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
+        buttons = self.get_widgets_in_contents(
+            filter_function=self.is_a(ui.MenuButton))
+        # A button is created for each existing environment (see details) and
+        # for each environment type supported by quickstart (create).
+        env_types = envs.get_supported_env_types(self.env_type_db)
+        expected_buttons_number = (
+            # The number of active environments.
+            len(jenv_db['environments']) +
+            # The buttons to create new environments.
+            len(env_types) +
+            # The button to automatically create a new local environment.
+            1
+        )
+        self.assertEqual(expected_buttons_number, len(buttons))
+        # A text widget is displayed as header for the imported environments.
+        widgets = self.get_widgets_in_contents(
+            filter_function=self.is_a(urwid.Text))
+        self.assertEqual(self.active_environments_message, widgets[2].text)
+
+    def test_view_contents_without_imported_envs(self):
+        # If there are no active imported environments the corresponding
+        # header text is not displayed.
+        env_db = jenv_db = {'environments': {}}
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
+        widgets = self.get_widgets_in_contents(
+            filter_function=self.is_a(urwid.Text))
+        texts = [widget.text for widget in widgets]
+        self.assertNotIn(self.active_environments_message, texts)
+
     def test_new_local_environment_disabled(self):
         # The option to create a new local environment is not present if they
         # are not supported in the current platform.
         env_db = helpers.make_env_db()
+        jenv_db = helpers.make_jenv_db()
         with local_envs_supported(False):
-            views.env_index(
-                self.app, self.env_type_db, env_db, self.save_callable)
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
         captions = map(cli_helpers.get_button_caption, buttons)
@@ -202,7 +326,9 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
     def test_environment_clicked(self, mock_env_detail):
         # The environment detail view is called when clicking an environment.
         env_db = helpers.make_env_db()
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = {'environments': {}}
+        params = self.make_params(env_db, jenv_db)
+        views.env_index(self.app, params)
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
         # The environments are listed in alphabetical order.
@@ -216,21 +342,47 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
             # When the button is clicked, the detail view is called passing the
             # corresponding environment data.
             cli_helpers.emit(button)
-            mock_env_detail.assert_called_once_with(
-                self.app, self.env_type_db, env_db, self.save_callable,
-                env_data)
+            mock_env_detail.assert_called_once_with(self.app, params, env_data)
             # Reset the mock so that it does not include any calls on the next
             # loop cycle.
             mock_env_detail.reset_mock()
 
+    @mock.patch('quickstart.cli.views.jenv_detail')
+    def test_imported_environment_clicked(self, mock_jenv_detail):
+        # The jenv detail view is called when clicking an imported environment.
+        env_db = {'environments': {}}
+        jenv_db = helpers.make_jenv_db()
+        params = self.make_params(env_db, jenv_db)
+        with local_envs_supported(False):
+            views.env_index(self.app, params)
+        buttons = self.get_widgets_in_contents(
+            filter_function=self.is_a(ui.MenuButton))
+        # The environments are listed in alphabetical order.
+        environments = sorted(jenv_db['environments'])
+        for env_name, button in zip(environments, buttons):
+            env_data = envs.get_env_data(jenv_db, env_name)
+            # The caption includes the environment description.
+            env_description = jenv.get_env_short_description(env_data)
+            self.assertIn(
+                env_description, cli_helpers.get_button_caption(button))
+            # When the button is clicked, the jenv detail view is called
+            # passing the corresponding environment data.
+            cli_helpers.emit(button)
+            mock_jenv_detail.assert_called_once_with(
+                self.app, params, env_data)
+            # Reset the mock so that it does not include any calls on the next
+            # loop cycle.
+            mock_jenv_detail.reset_mock()
+
     @mock.patch('quickstart.cli.views.env_edit')
     def test_create_new_environment_clicked(self, mock_env_edit):
         # The environment edit view is called when clicking to create a new
         # environment.
         env_db = helpers.make_env_db()
+        jenv_db = {'environments': {}}
+        params = self.make_params(env_db, jenv_db)
         with local_envs_supported(True):
-            views.env_index(
-                self.app, self.env_type_db, env_db, self.save_callable)
+            views.env_index(self.app, params)
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
         env_types = envs.get_supported_env_types(self.env_type_db)
@@ -244,9 +396,10 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
             # corresponding environment data.
             cli_helpers.emit(button)
             mock_env_edit.assert_called_once_with(
-                self.app, self.env_type_db, env_db, self.save_callable,
-                {'type': env_type,
-                 'default-series': settings.JUJU_GUI_SUPPORTED_SERIES[-1]})
+                self.app, params, {
+                    'type': env_type,
+                    'default-series': settings.JUJU_GUI_SUPPORTED_SERIES[-1],
+                })
             # Reset the mock so that it does not include any calls on the next
             # loop cycle.
             mock_env_edit.reset_mock()
@@ -257,9 +410,10 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
         # If that option is clicked, the view quits the application returning
         # the newly created env_data.
         env_db = envs.create_empty_env_db()
-        with local_envs_supported(True):
-            views.env_index(
-                self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        with maas_env_detected(False):
+            with local_envs_supported(True):
+                views.env_index(self.app, self.make_params(env_db, jenv_db))
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
         # The "create and bootstrap" button is the first one in the contents.
@@ -280,19 +434,59 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
         # environment is not displayed if the current platform does not support
         # local environments.
         env_db = envs.create_empty_env_db()
+        jenv_db = helpers.make_jenv_db()
         with local_envs_supported(False):
-            views.env_index(
-                self.app, self.env_type_db, env_db, self.save_callable)
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
         buttons = self.get_widgets_in_contents(
             filter_function=self.is_a(ui.MenuButton))
-        # No "create and bootstrap" buttons are present.
+        # No "create and bootstrap local" buttons are present.
         captions = map(cli_helpers.get_button_caption, buttons)
         self.assertNotIn(self.create_local_caption, captions)
 
+    def test_create_and_bootstrap_maas_environment_clicked(self):
+        # When there are no environments in the env_db and a remote MAAS API is
+        # detected, the view exposes an option to automatically create and
+        # bootstrap a new MAAS environment.
+        # If that option is clicked, the view quits the application returning
+        # the newly created env_data.
+        env_db = envs.create_empty_env_db()
+        jenv_db = helpers.make_jenv_db()
+        with maas_env_detected(True):
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
+        buttons = self.get_widgets_in_contents(
+            filter_function=self.is_a(ui.MenuButton))
+        # The "create and bootstrap" button is the first one in the contents.
+        create_button = buttons[0]
+        self.assertEqual(
+            self.create_maas_caption,
+            cli_helpers.get_button_caption(create_button))
+        # An AppExit is raised clicking the button.
+        with self.assertRaises(ui.AppExit) as context_manager:
+            cli_helpers.emit(create_button)
+        new_env_db, env_data = context_manager.exception.return_value
+        # The environments database is no longer empty.
+        self.assertIn(MAAS_NAME, new_env_db['environments'])
+        self.assertEqual(envs.get_env_data(new_env_db, MAAS_NAME), env_data)
+
+    def test_create_and_bootstrap_maas_environment_missing(self):
+        # The option to automatically create and bootstrap a new MAAS
+        # environment is not displayed if no MAAS API endpoints are
+        # available on the system
+        env_db = envs.create_empty_env_db()
+        jenv_db = helpers.make_jenv_db()
+        with maas_env_detected(False):
+            views.env_index(self.app, self.make_params(env_db, jenv_db))
+        buttons = self.get_widgets_in_contents(
+            filter_function=self.is_a(ui.MenuButton))
+        # No "create and bootstrap MAAS" buttons are present.
+        captions = map(cli_helpers.get_button_caption, buttons)
+        self.assertNotIn(self.create_maas_caption, captions)
+
     def test_selected_environment(self):
         # The default environment is already selected in the list.
         env_db = helpers.make_env_db(default='lxc')
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         env_data = envs.get_env_data(env_db, 'lxc')
         env_description = envs.get_env_short_description(env_data)
         contents = self.app.get_contents()
@@ -304,31 +498,45 @@ class TestEnvIndex(EnvViewTestsMixin, unittest.TestCase):
     def test_status_with_errors(self):
         # The status message explains how errors are displayed.
         env_db = helpers.make_env_db()
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = {'environments': {}}
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         status = self.app.get_status()
         self.assertEqual(self.base_status + ' \N{BULLET} has errors ', status)
 
     def test_status_with_default(self):
         # The status message explains how default environment is represented.
         env_db = helpers.make_env_db(default='lxc', exclude_invalid=True)
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = {'environments': {}}
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         status = self.app.get_status()
         self.assertEqual(self.base_status + ' \N{CHECK MARK} default ', status)
 
-    def test_status_with_default_and_errors(self):
-        # The status message includes both default and errors explanations.
+    def test_status_with_active(self):
+        # The status message explains how active environments are displayed.
+        env_db = helpers.make_env_db(exclude_invalid=True)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
+        status = self.app.get_status()
+        self.assertEqual(self.base_status + ' \N{BULLET} active ', status)
+
+    def test_complete_status(self):
+        # The status message includes default, active and errors explanations.
         env_db = helpers.make_env_db(default='lxc')
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = helpers.make_jenv_db()
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         status = self.app.get_status()
         self.assertEqual(
             self.base_status +
-            ' \N{CHECK MARK} default  \N{BULLET} has errors ',
+            ' \N{CHECK MARK} default ' +
+            ' \N{BULLET} active ' +
+            ' \N{BULLET} has errors ',
             status)
 
-    def test_status(self):
+    def test_base_status(self):
         # The status only includes navigation info if there are no errors.
         env_db = helpers.make_env_db(exclude_invalid=True)
-        views.env_index(self.app, self.env_type_db, env_db, self.save_callable)
+        jenv_db = {'environments': {}}
+        views.env_index(self.app, self.make_params(env_db, jenv_db))
         status = self.app.get_status()
         self.assertEqual(self.base_status, status)
 
@@ -337,13 +545,13 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
 
     base_status = ' \N{RIGHTWARDS ARROW OVER LEFTWARDS ARROW} navigate '
     env_db = helpers.make_env_db(default='lxc')
+    jenv_db = helpers.make_jenv_db()
 
     def call_view(self, env_name='lxc'):
         """Call the view passing the env_data corresponding to env_name."""
         self.env_data = envs.get_env_data(self.env_db, env_name)
-        return views.env_detail(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            self.env_data)
+        self.params = self.make_params(self.env_db, self.jenv_db)
+        return views.env_detail(self.app, self.params, self.env_data)
 
     def test_view_default_return_value_on_exit(self):
         # The view configures the app so that the return value on user exit is
@@ -409,8 +617,7 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
         # The "back" button is the first one.
         back_button = self.get_control_buttons()[0]
         cli_helpers.emit(back_button)
-        mock_env_index.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable)
+        mock_env_index.assert_called_once_with(self.app, self.params)
 
     def test_use_button(self):
         # The application exits if the "use" button is clicked.
@@ -432,14 +639,13 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
         # The "set default" button is the third one.
         set_default_button = self.get_control_buttons()[2]
         cli_helpers.emit(set_default_button)
-        # The index view has been called passing the modified env_db as third
-        # argument.
+        # The index view has been called passing the modified env_db in params.
         self.assertTrue(mock_env_index.called)
-        new_env_db = mock_env_index.call_args[0][2]
+        params = mock_env_index.call_args[0][1]
         # The new env_db has a new default.
-        self.assertEqual(new_env_db['default'], 'ec2-west')
+        self.assertEqual(params.env_db['default'], 'ec2-west')
         # The new env_db has been saved.
-        self.save_callable.assert_called_once_with(new_env_db)
+        self.save_callable.assert_called_once_with(params.env_db)
 
     @mock.patch('quickstart.cli.views.env_edit')
     def test_edit_button(self, mock_env_edit):
@@ -449,24 +655,12 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
         edit_button = self.get_control_buttons()[3]
         cli_helpers.emit(edit_button)
         mock_env_edit.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            self.env_data)
+            self.app, self.params, self.env_data)
 
     def test_remove_button(self):
         # A confirmation dialog is displayed if the "remove" button is clicked.
         self.call_view(env_name='ec2-west')
-        original_contents = self.app.get_contents()
-        # The "remove" button is the last one.
-        remove_button = self.get_control_buttons()[-1]
-        cli_helpers.emit(remove_button)
-        # The original env detail contents have been replaced.
-        contents = self.app.get_contents()
-        self.assertIsNot(contents, original_contents)
-        # A "remove" confirmation dialog is displayed.
-        title_widget, message_widget, buttons = cli_helpers.inspect_dialog(
-            contents)
-        self.assertEqual('Remove the ec2-west environment', title_widget.text)
-        self.assertEqual('This action cannot be undone!', message_widget.text)
+        buttons, _ = self.click_remove_button('ec2-west')
         # The dialog includes the "cancel" and "confirm" buttons.
         self.assertEqual(2, len(buttons))
         captions = map(cli_helpers.get_button_caption, buttons)
@@ -475,15 +669,7 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
     def test_remove_cancelled(self):
         # The "remove" confirmation dialog can be safely dismissed.
         self.call_view(env_name='ec2-west')
-        original_contents = self.app.get_contents()
-        # The "remove" button is the last one.
-        remove_button = self.get_control_buttons()[-1]
-        cli_helpers.emit(remove_button)
-        contents = self.app.get_contents()
-        buttons = cli_helpers.inspect_dialog(contents)[2]
-        # The "cancel" button is the first one in the dialog.
-        cancel_button = buttons[0]
-        cli_helpers.emit(cancel_button)
+        original_contents = self.cancel_removal('ec2-west')
         # The original contents have been restored.
         self.assertIs(original_contents, self.app.get_contents())
 
@@ -492,23 +678,19 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
         # The current environment is removed if the "remove" button is clicked
         # and then the deletion is confirmed. Subsequently the application
         # switches to the index view.
-        self.call_view(env_name='ec2-west')
-        # The "remove" button is the last one.
-        remove_button = self.get_control_buttons()[-1]
-        cli_helpers.emit(remove_button)
-        contents = self.app.get_contents()
-        buttons = cli_helpers.inspect_dialog(contents)[2]
-        # The "confirm" button is the second one in the dialog.
-        confirm_button = buttons[1]
-        cli_helpers.emit(confirm_button)
-        # The index view has been called passing the modified env_db as third
-        # argument.
+        env_name = 'ec2-west'
+        self.call_view(env_name=env_name)
+        self.confirm_removal(env_name)
+        # A message notifies the environment has been removed.
+        self.assertEqual(
+            '{} successfully removed'.format(env_name), self.app.get_message())
+        # The index view has been called passing the modified env_db in params.
         self.assertTrue(mock_env_index.called)
-        new_env_db = mock_env_index.call_args[0][2]
+        params = mock_env_index.call_args[0][1]
         # The new env_db no longer includes the "ec2-west" environment.
-        self.assertNotIn('ec2-west', new_env_db['environments'])
+        self.assertNotIn(env_name, params.env_db['environments'])
         # The new env_db has been saved.
-        self.save_callable.assert_called_once_with(new_env_db)
+        self.save_callable.assert_called_once_with(params.env_db)
 
     def test_status_with_errors(self):
         # The status message explains how field errors are displayed.
@@ -525,9 +707,132 @@ class TestEnvDetail(EnvViewTestsMixin, unittest.TestCase):
         self.assertEqual(self.base_status, status)
 
 
+class TestJenvDetail(EnvViewTestsMixin, unittest.TestCase):
+
+    env_db = helpers.make_env_db(default='lxc')
+    jenv_db = helpers.make_jenv_db()
+
+    def call_view(self, env_name='lxc'):
+        """Call the view passing the env_data corresponding to env_name."""
+        self.env_data = envs.get_env_data(self.jenv_db, env_name)
+        self.params = self.make_params(self.env_db, self.jenv_db)
+        return views.jenv_detail(self.app, self.params, self.env_data)
+
+    def test_view_default_return_value_on_exit(self):
+        # The view configures the app so that the return value on user exit is
+        # a tuple including a copy of the given env_db and None, the latter
+        # meaning no environment has been selected (for now).
+        self.call_view()
+        new_env_db, env_data = self.get_on_exit_return_value(self.loop)
+        self.assertEqual(self.env_db, new_env_db)
+        self.assertIsNot(self.env_db, new_env_db)
+        self.assertIsNone(env_data)
+
+    def test_view_title(self):
+        # The application title is correctly set up: it shows the description
+        # of the current jenv environment.
+        self.call_view()
+        env_description = jenv.get_env_short_description(self.env_data)
+        self.assertEqual(env_description, self.app.get_title())
+
+    def test_view_contents(self):
+        # The view displays the jenv details.
+        self.call_view()
+        widgets = self.get_widgets_in_contents(
+            filter_function=self.is_a(urwid.Text))
+        expected_texts = [
+            '{}: {}'.format(label, value) for label, value
+            in jenv.get_env_details(self.env_data)
+        ]
+        for expected_text, widget in zip(expected_texts, widgets):
+            self.assertEqual(expected_text, widget.text)
+
+    def test_view_buttons(self):
+        # The "back", "use" and "remove" buttons are displayed.
+        self.call_view(env_name='ec2-west')
+        buttons = self.get_control_buttons()
+        captions = map(cli_helpers.get_button_caption, buttons)
+        self.assertEqual(['back', 'use', 'remove'], captions)
+
+    @mock.patch('quickstart.cli.views.env_index')
+    def test_back_button(self, mock_env_index):
+        # The index view is called if the "back" button is clicked.
+        self.call_view(env_name='ec2-west')
+        # The "back" button is the first one.
+        back_button = self.get_control_buttons()[0]
+        cli_helpers.emit(back_button)
+        mock_env_index.assert_called_once_with(self.app, self.params)
+
+    def test_use_button(self):
+        # The application exits if the "use" button is clicked.
+        # The env_db and the current environment data are returned.
+        self.call_view(env_name='ec2-west')
+        # The "use" button is the second one.
+        use_button = self.get_control_buttons()[1]
+        with self.assertRaises(ui.AppExit) as context_manager:
+            cli_helpers.emit(use_button)
+        expected_return_value = (self.env_db, self.env_data)
+        self.assertEqual(
+            expected_return_value, context_manager.exception.return_value)
+
+    def test_remove_button(self):
+        # A confirmation dialog is displayed if the "remove" button is clicked.
+        self.call_view(env_name='test-jenv')
+        buttons, _ = self.click_remove_button('test-jenv')
+        # The dialog includes the "cancel" and "confirm" buttons.
+        self.assertEqual(2, len(buttons))
+        captions = map(cli_helpers.get_button_caption, buttons)
+        self.assertEqual(['cancel', 'confirm'], captions)
+
+    def test_remove_cancelled(self):
+        # The "remove" confirmation dialog can be safely dismissed.
+        self.call_view(env_name='test-jenv')
+        original_contents = self.cancel_removal('test-jenv')
+        # The original contents have been restored.
+        self.assertIs(original_contents, self.app.get_contents())
+
+    @mock.patch('quickstart.cli.views.env_index')
+    def test_remove_confirmed(self, mock_env_index):
+        # The jenv file is removed if the "remove" button is clicked and then
+        # then the deletion is confirmed. Subsequently the application switches
+        # to the index view.
+        env_name = 'test-jenv'
+        self.call_view(env_name=env_name)
+        self.confirm_removal(env_name)
+        # A message notifies the environment has been removed.
+        self.assertEqual(
+            '{} successfully removed'.format(env_name), self.app.get_message())
+        # The index view has been called passing the modified jenv_db params.
+        self.assertTrue(mock_env_index.called)
+        params = mock_env_index.call_args[0][1]
+        # The new jenv_db no longer includes the "test-jenv" environment.
+        self.assertNotIn(env_name, params.jenv_db['environments'])
+        # The corresponding jenv file has been removed.
+        self.remove_jenv_callable.assert_called_once_with(env_name)
+        self.assertEqual(
+            'test-jenv successfully removed', self.app.get_message())
+
+    @mock.patch('quickstart.cli.views.env_index')
+    def test_remove_confirmed_error(self, mock_env_index):
+        # Errors occurred while trying to remove the jenv files are notified.
+        env_name = 'test-jenv'
+        self.call_view(env_name=env_name)
+        # Simulate an error removing the jenv file.
+        self.remove_jenv_callable.return_value = 'bad wolf'
+        self.confirm_removal(env_name)
+        # The error is notified.
+        self.assertEqual('bad wolf'.format(env_name), self.app.get_message())
+        # The index view has been called passing the original jenv_db params.
+        self.assertTrue(mock_env_index.called)
+        params = mock_env_index.call_args[0][1]
+        # The jenv_db still includes the "test_jenv" environment.
+        self.assertIn(env_name, params.jenv_db['environments'])
+
+
 class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
 
     env_db = helpers.make_env_db(default='lxc')
+    jenv_db = helpers.make_jenv_db()
 
     def call_view(self, env_name='lxc', env_type=None):
         """Call the view passing the env_data corresponding to env_name.
@@ -539,9 +844,8 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
             self.env_data = envs.get_env_data(self.env_db, env_name)
         else:
             self.env_data = {'type': env_type}
-        return views.env_edit(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            self.env_data)
+        self.params = self.make_params(self.env_db, self.jenv_db)
+        return views.env_edit(self.app, self.params, self.env_data)
 
     def get_form_contents(self):
         """Return the form contents included in the app page.
@@ -633,14 +937,14 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
         expected_form_contents = [
             ('provider type: ', 'local'),
             ('environment name: ', 'lxc'),
-            ('admin secret: ', 'bones'),
-            ('default series: ', 'raring'),
+            ('use lxc-clone', forms.MIXED),
+            ('use lxc-clone-aufs', forms.MIXED),
             ('root dir: ', ''),
             ('storage port: ', '8888'),
             ('shared storage port: ', ''),
             ('network bridge: ', ''),
-            ('use lxc-clone', forms.MIXED),
-            ('use lxc-clone-aufs', forms.MIXED),
+            ('admin secret: ', 'bones'),
+            ('default series: ', 'raring'),
             ('default', True),
         ]
         self.assertEqual(expected_form_contents, self.get_form_contents())
@@ -676,8 +980,7 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
             'ec2-west successfully modified', self.app.get_message())
         # The application displays the environment detail view.
         mock_env_detail.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            new_env_data)
+            self.app, self.params, new_env_data)
 
     @mock.patch('quickstart.cli.views.env_detail')
     def test_save_empty_db(self, mock_env_detail):
@@ -696,8 +999,7 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
         expected_new_env_data.update({'type': 'local', 'is-default': True})
         envs.set_env_data(self.env_db, None, expected_new_env_data)
         mock_env_detail.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            expected_new_env_data)
+            self.app, self.params, expected_new_env_data)
 
     def test_save_invalid_form_data(self):
         # Errors are displayed if the user tries to save invalid data.
@@ -741,8 +1043,7 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
         # The "cancel" button is the second one.
         cancel_button = self.get_control_buttons()[1]
         cli_helpers.emit(cancel_button)
-        mock_env_index.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable)
+        mock_env_index.assert_called_once_with(self.app, self.params)
 
     @mock.patch('quickstart.cli.views.env_detail')
     def test_modification_view_cancel_button(self, mock_env_detail):
@@ -753,5 +1054,4 @@ class TestEnvEdit(EnvViewTestsMixin, unittest.TestCase):
         cancel_button = self.get_control_buttons()[1]
         cli_helpers.emit(cancel_button)
         mock_env_detail.assert_called_once_with(
-            self.app, self.env_type_db, self.env_db, self.save_callable,
-            self.env_data)
+            self.app, self.params, self.env_data)
