@@ -21,6 +21,7 @@ from __future__ import unicode_literals
 from contextlib import contextmanager
 import os
 import shutil
+import socket
 import tempfile
 
 import mock
@@ -36,8 +37,11 @@ def assert_logs(messages, level='debug'):
     """
     with mock.patch('logging.{}'.format(level.lower())) as mock_log:
         yield
-    expected_calls = [mock.call(message) for message in messages]
-    mock_log.assert_has_calls(expected_calls)
+    if messages:
+        expected_calls = [mock.call(message) for message in messages]
+        mock_log.assert_has_calls(expected_calls)
+    else:
+        assert not mock_log.called, 'logging unexpectedly called'
 
 
 class BundleFileTestsMixin(object):
@@ -133,6 +137,73 @@ class EnvFileTestsMixin(object):
         return env_file.name
 
 
+class JenvFileTestsMixin(object):
+    """Shared methods for testing Juju generated environment files (jenv)."""
+
+    jenv_data = {
+        'user': 'admin',
+        'password': 'Secret!',
+        'state-servers': ['localhost:17070', '10.0.3.1:17070'],
+        'bootstrap-config': {
+            'admin-secret': 'Secret!',
+            'api-port': 17070,
+            'type': 'ec2',
+        },
+        'life': {'universe': {'everything': 42}},
+    }
+
+    def _make_playground(self):
+        """Create and return a mock Juju home."""
+        playground = tempfile.mkdtemp()
+        self.addCleanup(shutil.rmtree, playground)
+        os.mkdir(os.path.join(playground, 'environments'))
+        return playground
+
+    def write_jenv(self, juju_home, name, contents):
+        """Create a jenv file in the given Juju home.
+
+        Use the given name and YAML encoded contents to create the jenv.
+        Return the resulting jenv path.
+        """
+        path = os.path.join(juju_home, 'environments', '{}.jenv'.format(name))
+        with open(path, 'w') as jenv_file:
+            jenv_file.write(contents)
+        return path
+
+    @contextmanager
+    def make_jenv(self, env_name, contents):
+        """Create a temporary jenv file with the given env_name and contents.
+
+        In the context manager block, the JUJU_HOME is set to the ancestor
+        of the generated temporary file.
+
+        Return the jenv file path.
+        """
+        playground = self._make_playground()
+        jenv_path = self.write_jenv(playground, env_name, contents)
+        # Patch the JUJU_HOME and return the jenv file path.
+        with mock.patch('quickstart.settings.JUJU_HOME', playground):
+            yield jenv_path
+
+    @contextmanager
+    def make_multiple_jenvs(self, data):
+        """Create multiple temporary jenv files with the data.
+
+        Data is a dict mapping env names to jenv contents.
+
+        In the context manager block, the JUJU_HOME is set to the ancestor
+        of the generated temporary files.
+
+        Return the Juju home path.
+        """
+        playground = self._make_playground()
+        for name, contents in data.items():
+            self.write_jenv(playground, name, contents)
+        # Patch the JUJU_HOME and return the jenv file path.
+        with mock.patch('quickstart.settings.JUJU_HOME', playground):
+            yield playground
+
+
 def make_env_db(default=None, exclude_invalid=False):
     """Create and return an env_db.
 
@@ -188,15 +259,60 @@ def make_env_db(default=None, exclude_invalid=False):
     return env_db
 
 
+def make_jenv_db():
+    """Create and return a jenv files database."""
+    environments = {
+        'ec2-west': {
+            'type': '__unknown__',
+            'user': 'who',
+            'state-servers': ('1.2.3.4:42', '1.2.3.4:47'),
+        },
+        'lxc': {
+            'type': 'local',
+            'user': 'dalek',
+            'state-servers': ('localhost:17070', '10.0.3.1:17070'),
+        },
+        'test-jenv': {
+            'type': '__unknown__',
+            'user': 'my-user',
+            'state-servers': ('10.0.3.1:17070',),
+        },
+    }
+    return {'environments': environments}
+
+
 # Mock the builtin print function.
 mock_print = mock.patch('__builtin__.print')
 
 
+def patch_socket_create_connection(error=None):
+    """Patch the socket.create_connection function.
+
+    If error is not None, the mock object raises a socket.error with the given
+    message.
+    """
+    mock_create_connection = mock.Mock()
+    if error is not None:
+        mock_create_connection.side_effect = socket.error(error)
+    return mock.patch('socket.create_connection', mock_create_connection)
+
+
+def patch_check_resolvable(error=None):
+    """Patch the netutils.check_resolvable function to return the given error.
+
+    This is done so that tests do not try to resolve hostname addresses.
+    """
+    return mock.patch(
+        'quickstart.netutils.check_resolvable',
+        lambda hostname: error,
+    )
+
+
 class UrlReadTestsMixin(object):
-    """Expose a method to mock the quickstart.utils.urlread helper function."""
+    """Helpers to mock the quickstart.netutils.urlread helper function."""
 
     def patch_urlread(self, contents=None, error=False):
-        """Patch the quickstart.utils.urlread helper function.
+        """Patch the quickstart.netutils.urlread helper function.
 
         If contents is not None, urlread() will return the provided contents.
         If error is set to True, an IOError will be simulated.
@@ -206,7 +322,7 @@ class UrlReadTestsMixin(object):
             mock_urlread.return_value = contents
         if error:
             mock_urlread.side_effect = IOError('bad wolf')
-        return mock.patch('quickstart.utils.urlread', mock_urlread)
+        return mock.patch('quickstart.netutils.urlread', mock_urlread)
 
 
 class ValueErrorTestsMixin(object):
