@@ -99,28 +99,6 @@ class TestCreateEmptyEnvDb(unittest.TestCase):
         self.assertEqual({'environments': {}}, env_db)
 
 
-class TestLoadFile(
-        helpers.EnvFileTestsMixin, helpers.ValueErrorTestsMixin,
-        unittest.TestCase):
-
-    def test_no_file(self):
-        # A ValueError is raised if the environments file is not found.
-        expected = (
-            "unable to open environments file: "
-            "[Errno 2] No such file or directory: '/no/such/file.yaml'"
-        )
-        with self.assert_value_error(expected):
-            envs._load_file('/no/such/file.yaml')
-
-    def test_invalid_yaml(self):
-        # A ValueError is raised if the environments file is not a valid YAML.
-        env_file = self.make_env_file(':')
-        with self.assertRaises(ValueError) as context_manager:
-            envs._load_file(env_file)
-        expected = 'unable to parse environments file {}'.format(env_file)
-        self.assertIn(expected, bytes(context_manager.exception))
-
-
 class TestLoad(
         helpers.EnvFileTestsMixin, helpers.ValueErrorTestsMixin,
         unittest.TestCase):
@@ -205,60 +183,6 @@ class TestLoad(
         self.assertEqual(expected, env_db)
 
 
-class TestLoadGenerated(
-        helpers.EnvFileTestsMixin, helpers.ValueErrorTestsMixin,
-        unittest.TestCase):
-
-    def test_empty_file(self):
-        # A ValueError is raised if the environments file is empty.
-        env_file = self.make_env_file('')
-        expected = 'invalid YAML contents in {}: None'.format(env_file)
-        with self.assert_value_error(expected):
-            envs.load_generated(env_file)
-
-    def test_invalid_yaml_contents(self):
-        # A ValueError is raised if the environments file is not well formed.
-        env_file = self.make_env_file('a-string')
-        expected = 'invalid YAML contents in {}: a-string'.format(env_file)
-        with self.assert_value_error(expected):
-            envs.load_generated(env_file)
-
-    def test_section_not_found(self):
-        expected = {
-            'shoehorn-config': {
-                'admin-secret': 'Secret!',
-                'type': 'ec2'},
-        }
-        yaml_contents = expected.copy()
-        env_file = self.make_env_file(yaml.safe_dump(yaml_contents))
-        expected = 'invalid YAML contents in {}: {}'.format(
-            env_file, yaml_contents)
-        with self.assert_value_error(expected):
-            envs.load_generated(env_file)
-
-    def test_successful_default_section(self):
-        expected = {
-            'bootstrap-config': {
-                'admin-secret': 'Secret!',
-                'type': 'ec2'},
-        }
-        yaml_contents = expected.copy()
-        env_file = self.make_env_file(yaml.safe_dump(yaml_contents))
-        env_db = envs.load_generated(env_file)
-        self.assertEqual(expected['bootstrap-config'], env_db)
-
-    def test_successful_specified_section(self):
-        expected = {
-            'my-config': {
-                'admin-secret': 'Secret!',
-                'type': 'ec2'},
-        }
-        yaml_contents = expected.copy()
-        env_file = self.make_env_file(yaml.safe_dump(yaml_contents))
-        env_db = envs.load_generated(env_file, section='my-config')
-        self.assertEqual(expected['my-config'], env_db)
-
-
 class TestSave(helpers.EnvFileTestsMixin, unittest.TestCase):
 
     def setUp(self):
@@ -644,6 +568,41 @@ class TestCreateLocalEnvData(unittest.TestCase):
         self.assertEqual(expected_env_data, env_data)
 
 
+class TestCreateMaasEnvData(unittest.TestCase):
+
+    def setUp(self):
+        # Store the env_type_db.
+        self.env_type_db = envs.get_env_type_db()
+
+    def test_not_default(self):
+        # The resulting env_data is correctly structured for non default envs.
+        env_data = envs.create_maas_env_data(
+            self.env_type_db, 'my-maas', 'http://1.2.3.4/MAAS', 'Secret!',
+            is_default=False)
+        expected_env_data = {
+            'type': 'maas',
+            'name': 'my-maas',
+            'maas-server': 'http://1.2.3.4/MAAS',
+            'maas-oauth': 'Secret!',
+            'is-default': False,
+        }
+        self.assertEqual(expected_env_data, env_data)
+
+    def test_default(self):
+        # The resulting env_data is correctly structured for default envs.
+        env_data = envs.create_maas_env_data(
+            self.env_type_db, 'another-maas', 'http://1.2.3.4/MAAS', 'Boo!',
+            is_default=True)
+        expected_env_data = {
+            'type': 'maas',
+            'name': 'another-maas',
+            'maas-server': 'http://1.2.3.4/MAAS',
+            'maas-oauth': 'Boo!',
+            'is-default': True,
+        }
+        self.assertEqual(expected_env_data, env_data)
+
+
 class TestRemoveEnv(
         EnvDataTestsMixin, helpers.ValueErrorTestsMixin, unittest.TestCase):
 
@@ -740,10 +699,9 @@ class TestGetEnvTypeDb(unittest.TestCase):
         self.assertIn('local', self.env_type_db)
         env_metadata = self.env_type_db['local']
         expected = [
-            'type', 'name', 'admin-secret', 'default-series', 'root-dir',
+            'type', 'name', 'lxc-clone', 'lxc-clone-aufs', 'root-dir',
             'storage-port', 'shared-storage-port', 'network-bridge',
-            'lxc-clone', 'lxc-clone-aufs',
-            'is-default']
+            'admin-secret', 'default-series', 'is-default']
         expected_required = ['type', 'name', 'is-default']
         self.assert_fields(expected, env_metadata)
         self.assert_required_fields(expected_required, env_metadata)
@@ -753,8 +711,8 @@ class TestGetEnvTypeDb(unittest.TestCase):
         self.assertIn('ec2', self.env_type_db)
         env_metadata = self.env_type_db['ec2']
         expected = [
-            'type', 'name', 'admin-secret', 'default-series', 'access-key',
-            'secret-key', 'control-bucket', 'region', 'is-default']
+            'type', 'name', 'access-key', 'secret-key', 'region',
+            'admin-secret', 'default-series', 'control-bucket',  'is-default']
         expected_required = [
             'type', 'name', 'access-key', 'secret-key',
             'is-default']
@@ -766,13 +724,13 @@ class TestGetEnvTypeDb(unittest.TestCase):
         self.assertIn('openstack', self.env_type_db)
         env_metadata = self.env_type_db['openstack']
         expected = [
-            'type', 'name', 'admin-secret', 'default-series',
-            'use-floating-ip', 'control-bucket', 'auth-url', 'tenant-name',
+            'type', 'name', 'auth-url', 'tenant-name', 'use-floating-ip',
             'region', 'auth-mode', 'username', 'password', 'access-key',
-            'secret-key', 'is-default']
+            'secret-key', 'control-bucket', 'admin-secret', 'default-series',
+            'is-default']
         expected_required = [
-            'type', 'name', 'use-floating-ip',
-            'auth-url', 'tenant-name', 'region', 'is-default']
+            'type', 'name', 'auth-url', 'tenant-name', 'use-floating-ip',
+            'region', 'is-default']
         self.assert_fields(expected, env_metadata)
         self.assert_required_fields(expected_required, env_metadata)
 
@@ -781,13 +739,13 @@ class TestGetEnvTypeDb(unittest.TestCase):
         self.assertIn('azure', self.env_type_db)
         env_metadata = self.env_type_db['azure']
         expected = [
-            'type', 'name', 'admin-secret', 'default-series', 'location',
-            'management-subscription-id', 'management-certificate-path',
-            'storage-account-name', 'is-default']
+            'type', 'name', 'management-subscription-id',
+            'management-certificate-path', 'storage-account-name', 'location',
+            'admin-secret', 'default-series', 'is-default']
         expected_required = [
-            'type', 'name', 'location',
-            'management-subscription-id', 'management-certificate-path',
-            'storage-account-name', 'is-default']
+            'type', 'name', 'management-subscription-id',
+            'management-certificate-path', 'storage-account-name', 'location',
+            'is-default']
         self.assert_fields(expected, env_metadata)
         self.assert_required_fields(expected_required, env_metadata)
 
@@ -796,16 +754,40 @@ class TestGetEnvTypeDb(unittest.TestCase):
         self.assertIn('joyent', self.env_type_db)
         env_metadata = self.env_type_db['joyent']
         expected = [
-            'type', 'name', 'admin-secret', 'default-series',
-            'sdc-user', 'sdc-key-id', 'sdc-url',
-            'manta-user', 'manta-key-id', 'manta-url',
-            'private-key-path', 'algorithm', 'is-default']
+            'type', 'name', 'sdc-user', 'sdc-key-id', 'sdc-url', 'manta-user',
+            'manta-key-id', 'manta-url', 'private-key-path', 'algorithm',
+            'admin-secret', 'default-series', 'is-default']
         expected_required = [
             'type', 'name', 'sdc-user', 'sdc-key-id', 'manta-user',
             'manta-key-id', 'is-default']
         self.assert_fields(expected, env_metadata)
         self.assert_required_fields(expected_required, env_metadata)
 
+    def test_maas_environment(self):
+        # The MAAS environment metadata includes the expected fields.
+        self.assertIn('maas', self.env_type_db)
+        env_metadata = self.env_type_db['maas']
+        expected = [
+            'type', 'name', 'maas-server', 'maas-oauth',
+            'authorized-keys-path', 'admin-secret', 'default-series',
+            'is-default']
+        expected_required = [
+            'type', 'name', 'maas-server', 'maas-oauth', 'is-default']
+        self.assert_fields(expected, env_metadata)
+        self.assert_required_fields(expected_required, env_metadata)
+
+    def test_manual_environment(self):
+        # The manual environment metadata includes the expected fields.
+        self.assertIn('manual', self.env_type_db)
+        env_metadata = self.env_type_db['manual']
+        expected = [
+            'type', 'name', 'bootstrap-host', 'bootstrap-user',
+            'storage-listen-ip', 'storage-port', 'admin-secret', 'is-default']
+        expected_required = [
+            'type', 'name', 'bootstrap-host', 'is-default']
+        self.assert_fields(expected, env_metadata)
+        self.assert_required_fields(expected_required, env_metadata)
+
 
 class TestGetSupportedEnvTypes(unittest.TestCase):
 
@@ -820,6 +802,8 @@ class TestGetSupportedEnvTypes(unittest.TestCase):
             ('openstack', 'OpenStack (or HP Public Cloud)'),
             ('azure', 'Windows Azure'),
             ('joyent', 'Joyent'),
+            ('maas', 'MAAS (bare metal)'),
+            ('manual', 'Manual Provisioning'),
             ('local', 'local (LXC)'),
         ]
         obtained_env_types = envs.get_supported_env_types(self.env_type_db)
@@ -879,14 +863,14 @@ class TestMapFieldsToEnvData(unittest.TestCase):
         return [
             ('type', 'local'),
             ('name', 'lxc'),
-            ('admin-secret', 'Secret!'),
-            ('default-series', 'saucy'),
+            ('lxc-clone', False),
+            ('lxc-clone-aufs', False),
             ('root-dir', '/my/juju/local/'),
             ('storage-port', 4242),
             ('shared-storage-port', 4747),
             ('network-bridge', 'lxcbr1'),
-            ('lxc-clone', False),
-            ('lxc-clone-aufs', False),
+            ('admin-secret', 'Secret!'),
+            ('default-series', 'saucy'),
             ('is-default', True),
         ]
 
@@ -901,14 +885,14 @@ class TestMapFieldsToEnvData(unittest.TestCase):
         expected = [
             ('type', 'local'),
             ('name', 'lxc'),
-            ('admin-secret', None),
-            ('default-series', None),
+            ('lxc-clone', None),
+            ('lxc-clone-aufs', None),
             ('root-dir', None),
             ('storage-port', None),
             ('shared-storage-port', None),
             ('network-bridge', None),
-            ('lxc-clone', None),
-            ('lxc-clone-aufs', None),
+            ('admin-secret', None),
+            ('default-series', None),
             ('is-default', None),
         ]
         env_data = {'type': 'local', 'name': 'lxc'}
