@@ -35,10 +35,15 @@ from quickstart import (
     manage,
     settings,
 )
-from quickstart.cli import views
-from quickstart.models import envs
+from quickstart.cli import (
+    params,
+    views,
+)
+from quickstart.models import (
+    envs,
+    jenv,
+)
 from quickstart.tests import helpers
-from quickstart import app
 
 
 class TestDescriptionAction(unittest.TestCase):
@@ -373,13 +378,15 @@ class TestStartInteractiveSession(
         self.env_type_db = envs.get_env_type_db()
         self.env_file = self.make_env_file()
         self.env_db = envs.load(self.env_file)
+        self.jenv_db = helpers.make_jenv_db()
 
     @contextmanager
-    def patch_interactive_mode(self, env_db, return_value):
+    def patch_interactive_mode(self, env_db, jenv_db, return_value):
         """Patch the quickstart.cli.views.show function.
 
         Ensure the interactive mode is started by the code in the context block
-        passing the given env_db. Make the view return the given return_value.
+        passing the given env_db and jenv_db.
+        Make the view return the given return_value.
         """
         create_save_callable_path = 'quickstart.manage._create_save_callable'
         mock_show = mock.Mock(return_value=return_value)
@@ -387,17 +394,24 @@ class TestStartInteractiveSession(
             with mock.patch('quickstart.manage.views.show', mock_show):
                 yield
         mock_save_callable.assert_called_once_with(self.parser, self.env_file)
-        mock_show.assert_called_once_with(
-            views.env_index, self.env_type_db, env_db,
-            mock_save_callable())
+        expected_params = params.Params(
+            env_type_db=self.env_type_db,
+            env_db=env_db,
+            jenv_db=jenv_db,
+            save_callable=mock_save_callable(),
+            remove_jenv_callable=jenv.remove,
+        )
+        mock_show.assert_called_once_with(views.env_index, expected_params)
 
     def test_resulting_env_data(self):
         # The interactive session can be used to select an environment, in
         # which case the function returns the corresponding env_data.
         env_data = envs.get_env_data(self.env_db, 'aws')
-        with self.patch_interactive_mode(self.env_db, [self.env_db, env_data]):
+        with self.patch_interactive_mode(
+                self.env_db, self.jenv_db, [self.env_db, env_data]):
             obtained_env_data = manage._start_interactive_session(
-                self.parser, self.env_type_db, self.env_db, self.env_file)
+                self.parser, self.env_type_db, self.env_db, self.jenv_db,
+                self.env_file)
         self.assertEqual(env_data, obtained_env_data)
 
     @helpers.mock_print
@@ -406,9 +420,11 @@ class TestStartInteractiveSession(
         # during the interactive session.
         env_data = envs.get_env_data(self.env_db, 'aws')
         new_env_db = helpers.make_env_db()
-        with self.patch_interactive_mode(self.env_db, [new_env_db, env_data]):
+        with self.patch_interactive_mode(
+                self.env_db, self.jenv_db, [new_env_db, env_data]):
             manage._start_interactive_session(
-                self.parser, self.env_type_db, self.env_db, self.env_file)
+                self.parser, self.env_type_db, self.env_db, self.jenv_db,
+                self.env_file)
         mock_print.assert_called_once_with(
             'changes to the environments file have been saved')
 
@@ -416,9 +432,11 @@ class TestStartInteractiveSession(
     def test_interactive_mode_quit(self, mock_exit):
         # If the user explicitly quits the interactive mode, the program exits
         # without proceeding with the environment bootstrapping.
-        with self.patch_interactive_mode(self.env_db, [self.env_db, None]):
+        with self.patch_interactive_mode(
+                self.env_db, self.jenv_db, [self.env_db, None]):
             manage._start_interactive_session(
-                self.parser, self.env_type_db, self.env_db, self.env_file)
+                self.parser, self.env_type_db, self.env_db, self.jenv_db,
+                self.env_file)
         mock_exit.assert_called_once_with('quitting')
 
 
@@ -429,26 +447,37 @@ class TestRetrieveEnvData(unittest.TestCase):
         self.parser = mock.Mock()
         self.env_type_db = envs.get_env_type_db()
         self.env_db = helpers.make_env_db()
+        self.jenv_db = helpers.make_jenv_db()
 
     def test_resulting_env_data(self):
         # The env_data is correctly validated and returned.
         expected_env_data = envs.get_env_data(self.env_db, 'lxc')
         env_data = manage._retrieve_env_data(
-            self.parser, self.env_type_db, self.env_db, 'lxc')
+            self.parser, self.env_type_db, self.env_db, self.jenv_db, 'lxc')
+        self.assertEqual(expected_env_data, env_data)
+
+    def test_jenv_data(self):
+        # The env_data is correctly retrieved from the jenv database.
+        expected_env_data = envs.get_env_data(self.jenv_db, 'test-jenv')
+        env_data = manage._retrieve_env_data(
+            self.parser, self.env_type_db, self.env_db, self.jenv_db,
+            'test-jenv')
         self.assertEqual(expected_env_data, env_data)
 
     def test_error_environment_not_found(self):
         # A parser error is invoked if the provided environment is not included
         # in the environments database.
         manage._retrieve_env_data(
-            self.parser, self.env_type_db, self.env_db, 'no-such')
+            self.parser, self.env_type_db, self.env_db, self.jenv_db,
+            'no-such')
         self.parser.error.assert_called_once_with(
             'environment no-such not found')
 
     def test_error_environment_not_valid(self):
         # A parser error is invoked if the selected environment is not valid.
         manage._retrieve_env_data(
-            self.parser, self.env_type_db, self.env_db, 'local-with-errors')
+            self.parser, self.env_type_db, self.env_db, self.jenv_db,
+            'local-with-errors')
         self.parser.error.assert_called_once_with(
             'cannot use the local-with-errors environment:\n'
             'the storage port field requires an integer value')
@@ -478,7 +507,9 @@ class TestValidatePlatform(unittest.TestCase):
         self.assertIsNone(result)
 
 
-class TestSetupEnv(helpers.EnvFileTestsMixin, unittest.TestCase):
+class TestSetupEnv(
+        helpers.EnvFileTestsMixin, helpers.JenvFileTestsMixin,
+        unittest.TestCase):
 
     def setUp(self):
         self.parser = mock.Mock()
@@ -509,7 +540,6 @@ class TestSetupEnv(helpers.EnvFileTestsMixin, unittest.TestCase):
         options = self.make_options(
             env_file, env_name='aws', interactive=False)
         manage._setup_env(options, self.parser)
-        self.assertEqual('Secret!', options.admin_secret)
         self.assertEqual(env_file, options.env_file)
         self.assertEqual('aws', options.env_name)
         self.assertEqual('ec2', options.env_type)
@@ -550,7 +580,6 @@ class TestSetupEnv(helpers.EnvFileTestsMixin, unittest.TestCase):
         options = self.make_options(
             env_file, env_name='lxc', interactive=False)
         manage._setup_env(options, self.parser)
-        self.assertEqual('Secret!', options.admin_secret)
         self.assertEqual(env_file, options.env_file)
         self.assertEqual('lxc', options.env_name)
         self.assertEqual('local', options.env_type)
@@ -588,12 +617,13 @@ class TestSetupEnv(helpers.EnvFileTestsMixin, unittest.TestCase):
         env_data = envs.get_env_data(env_db, 'aws')
         get_env_type_db_path = 'quickstart.models.envs.get_env_type_db'
         with mock.patch(get_env_type_db_path) as mock_get_env_type_db:
-            with self.patch_interactive_mode(env_data) as mock_interactive:
-                manage._setup_env(options, self.parser)
+            with self.make_jenv('ec2', yaml.safe_dump(self.jenv_data)):
+                jenv_db = jenv.get_env_db()
+                with self.patch_interactive_mode(env_data) as mock_interactive:
+                    manage._setup_env(options, self.parser)
         mock_interactive.assert_called_once_with(
-            self.parser, mock_get_env_type_db(), env_db, env_file)
+            self.parser, mock_get_env_type_db(), env_db, jenv_db, env_file)
         # The options is updated with data from the selected environment.
-        self.assertEqual('Secret!', options.admin_secret)
         self.assertEqual(env_file, options.env_file)
         self.assertEqual('aws', options.env_name)
         self.assertEqual('ec2', options.env_type)
@@ -754,7 +784,6 @@ class TestRun(unittest.TestCase):
     def make_options(self, **kwargs):
         """Set up the options to be passed to the run function."""
         options = {
-            'admin_secret': 'Secretz!',
             'bundle': None,
             'bundle_id': None,
             'charm_url': None,
@@ -767,242 +796,178 @@ class TestRun(unittest.TestCase):
         options.update(kwargs)
         return mock.Mock(**options)
 
-    @staticmethod
-    def mock_get_value_from_jenv_success(name, home, key):
-        return 'from jenv'
+    def configure_app(self, mock_app, **kwargs):
+        """Configure the given mock_app with the given kwargs.
 
-    @staticmethod
-    def mock_get_value_from_jenv_error(name, home, key):
-        fn = '{}.jenv'.format(name)
-        path = os.path.join(home, 'environments', fn)
-        msg = '{} not found in {}'.format(key, path)
-        raise ValueError(msg.encode('utf-8'))
+        Each key/value pair in kwargs represents a mock_app attribute and
+        the associated return value. Those values are used to override
+        defaults.
 
-    def test_no_bundle(self, mock_app, mock_open):
+        Return the mock Juju environment object.
+        """
+        env = mock.Mock()
+        defaults = {
+            # Dependencies are installed.
+            'ensure_dependencies': (1, 18, 0),
+            # Ensure the current Juju version is supported.
+            'check_juju_supported': None,
+            # Ensure the SSH keys are properly configured.
+            'ensure_ssh_keys': None,
+            # The environment is not already bootstrapped.
+            'check_bootstrapped': None,
+            # This is also confirmed by the bootstrap function.
+            'bootstrap': False,
+            # Status is then called, returning the bootstrap node series.
+            'status': 'trusty',
+            # The API URL must be retrieved (the environment was not ready).
+            'get_api_url': 'wss://1.2.3.4:17070',
+            # Retrieve the environment credentials.
+            'get_credentials': ('MyUser', 'Secret!'),
+            # Connect to the Juju Environment API endpoint.
+            'connect': env,
+            # The environment is then checked.
+            'check_environment': (
+                'cs:trusty/juju-gui-42',
+                '0',
+                {'Name': 'juju-gui'},
+                {'Name': 'juju-gui/0'}
+            ),
+            # Deploy the Juju GUI charm.
+            'deploy_gui': 'juju-gui/0',
+            # Watch the deployment progress and return the unit address.
+            'watch': '1.2.3.4',
+            # Create the login token for the Juju GUI.
+            'create_auth_token': 'TOKEN',
+        }
+        defaults.update(kwargs)
+        for attr, return_value in defaults.items():
+            getattr(mock_app, attr).return_value = return_value
+        return env
+
+    def patch_get_juju_command(self):
+        """Patch the platform_support.get_juju_command function."""
+        path = 'quickstart.manage.platform_support.get_juju_command'
+        return mock.patch(path, return_value=(self.juju_command, False))
+
+    def test_run(self, mock_app, mock_open):
         # The application runs correctly if no bundle is provided.
-        mock_app.ensure_dependencies.return_value = (1, 18, 0)
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'trusty')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        service_data = {'Name': 'juju-gui'}
-        unit_data = {'Name': 'juju-gui/0'}
-        mock_app.check_environment.return_value = (
-            'cs:trusty/juju-gui-42', '0', service_data, unit_data)
-        mock_app.get_value_from_jenv = self.mock_get_value_from_jenv_error
-        # Make mock_app.watch return the Juju GUI unit address.
-        mock_app.watch.return_value = '1.2.3.4'
-        # Make mock_app.create_auth_token return a fake authentication token.
-        mock_app.create_auth_token.return_value = 'AUTHTOKEN'
+        env = self.configure_app(mock_app)
+        # Run the application.
         options = self.make_options()
-        with mock.patch('quickstart.manage.platform_support.get_juju_command',
-                        side_effect=[self.juju_command]):
+        with self.patch_get_juju_command():
             manage.run(options)
-        mock_app.ensure_dependencies.assert_called()
-        mock_app.ensure_ssh_keys.assert_called()
+        # Ensure the functions have been used correctly.
+        mock_app.ensure_dependencies.assert_called_once_with(
+            options.distro_only, options.platform, self.juju_command)
+        mock_app.check_juju_supported.assert_called_once_with((1, 18, 0))
+        mock_app.ensure_ssh_keys.assert_called_once_with()
+        mock_app.check_bootstrapped.assert_called_once_with(options.env_name)
         mock_app.bootstrap.assert_called_once_with(
-            options.env_name, self.juju_command, requires_sudo=False,
-            debug=options.debug, upload_tools=options.upload_tools,
+            options.env_name, self.juju_command,
+            debug=options.debug,
+            upload_tools=options.upload_tools,
+            upload_series=options.upload_series,
+            constraints=options.constraints)
+        mock_app.status.assert_called_once_with(
+            options.env_name, self.juju_command)
+        mock_app.get_api_url.assert_called_once_with(
+            options.env_name, self.juju_command)
+        mock_app.get_credentials.assert_called_once_with(options.env_name)
+        mock_app.connect.assert_has_calls([
+            mock.call('wss://1.2.3.4:17070', 'MyUser', 'Secret!'),
+            mock.call().close(),
+            mock.call('wss://1.2.3.4:443/ws', 'MyUser', 'Secret!'),
+            mock.call().close(),
+        ])
+        mock_app.check_environment.assert_called_once_with(
+            env, settings.JUJU_GUI_SERVICE_NAME, options.charm_url,
+            options.env_type, 'trusty', False)
+        mock_app.deploy_gui.assert_called_once_with(
+            env, settings.JUJU_GUI_SERVICE_NAME, 'cs:trusty/juju-gui-42',
+            '0', {'Name': 'juju-gui'}, {'Name': 'juju-gui/0'})
+        mock_app.watch.assert_called_once_with(env, 'juju-gui/0')
+        mock_app.create_auth_token.assert_called_once_with(env)
+        mock_open.assert_called_once_with('https://1.2.3.4/?authtoken=TOKEN')
+        # Ensure some of the app function have not been called.
+        self.assertFalse(mock_app.get_env_type.called)
+        self.assertFalse(mock_app.deploy_bundle.called)
+
+    def test_already_bootstrapped(self, mock_app, mock_open):
+        # The application correctly reuses an already bootstrapped environment.
+        env = self.configure_app(
+            mock_app, check_bootstrapped='wss://example.com')
+        # Run the application.
+        options = self.make_options()
+        with self.patch_get_juju_command():
+            manage.run(options)
+        # The environment type is retrieved from the jenv.
+        mock_app.get_env_type.assert_called_once_with(env)
+        # No reason to call bootstrap or get_api_url functions.
+        self.assertFalse(mock_app.bootstrap.called)
+        self.assertFalse(mock_app.get_api_url.called)
+
+    def test_already_bootstrapped_race(self, mock_app, mock_open):
+        # The application correctly reuses an already bootstrapped environment.
+        # In this case, the environment seems not bootstrapped at first, but
+        # it ended up being up and running later.
+        env = self.configure_app(mock_app, bootstrap=True)
+        # Run the application.
+        options = self.make_options()
+        with self.patch_get_juju_command():
+            manage.run(options)
+        # The bootstrap and get_api_url functions are still called, but this
+        # time also get_env_type is required.
+        # The environment type is retrieved from the jenv.
+        mock_app.bootstrap.assert_called_once_with(
+            options.env_name, self.juju_command,
+            debug=options.debug,
+            upload_tools=options.upload_tools,
             upload_series=options.upload_series,
             constraints=options.constraints)
         mock_app.get_api_url.assert_called_once_with(
             options.env_name, self.juju_command)
-        mock_app.connect.assert_has_calls([
-            mock.call(
-                mock_app.get_api_url(), options.admin_secret),
-            mock.call().close(),
-            mock.call('wss://1.2.3.4:443/ws', options.admin_secret),
-            mock.call().close(),
-        ])
-        mock_app.check_environment.assert_called_once_with(
-            mock_app.connect(), settings.JUJU_GUI_SERVICE_NAME,
-            options.charm_url, options.env_type, mock_app.bootstrap()[1],
-            mock_app.bootstrap()[0])
-        mock_app.deploy_gui.assert_called_once_with(
-            mock_app.connect(), settings.JUJU_GUI_SERVICE_NAME,
-            'cs:trusty/juju-gui-42', '0', service_data, unit_data)
-        mock_app.watch.assert_called_once_with(
-            mock_app.connect(), mock_app.deploy_gui())
-        mock_app.create_auth_token.assert_called_once_with(mock_app.connect())
-        mock_open.assert_called_once_with(
-            'https://{}/?authtoken={}'.format(mock_app.watch(), 'AUTHTOKEN'))
-        self.assertFalse(mock_app.deploy_bundle.called)
+        mock_app.get_env_type.assert_called_once_with(env)
 
     def test_no_token(self, mock_app, mock_open):
         # The process continues even if the authentication token cannot be
         # retrieved.
-        mock_app.create_auth_token.return_value = None
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
+        env = self.configure_app(mock_app, create_auth_token=None)
+        # Run the application.
         options = self.make_options()
-        manage.run(options)
-        mock_app.create_auth_token.assert_called_once_with(mock_app.connect())
-        mock_open.assert_called_once_with(
-            'https://{}'.format(mock_app.watch()))
+        with self.patch_get_juju_command():
+            manage.run(options)
+        # Ensure the browser is still open without an auth token.
+        mock_app.create_auth_token.assert_called_once_with(env)
+        mock_open.assert_called_once_with('https://1.2.3.4')
 
     def test_bundle(self, mock_app, mock_open):
         # A bundle is correctly deployed by the application.
+        env = self.configure_app(mock_app, create_auth_token=None)
+        # Run the application.
         options = self.make_options(
             bundle='/my/bundle/file.yaml', bundle_yaml='mybundle: contents',
             bundle_name='mybundle', bundle_services=['service1', 'service2'])
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'trusty')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:trusty/juju-gui-42', '0', None, None)
-        # Make mock_app.watch return the Juju GUI unit address.
-        mock_app.watch.return_value = 'gui.example.com'
-        manage.run(options)
+        with self.patch_get_juju_command():
+            manage.run(options)
+        # Ensure the bundle is correctly deployed.
         mock_app.deploy_bundle.assert_called_once_with(
-            mock_app.connect(), 'mybundle: contents', 'mybundle', None)
+            env, 'mybundle: contents', 'mybundle', None)
 
-    def test_local_provider_no_sudo(self, mock_app, mock_open):
+    def test_local_provider(self, mock_app, mock_open):
         # The application correctly handles working with local providers with
         # new Juju versions not requiring "sudo" to bootstrap the environment.
-        # Sudo privileges are not required if the Juju version is >= 1.17.2.
+        self.configure_app(mock_app, create_auth_token=None)
+        # Run the application.
         options = self.make_options(env_type='local')
-        versions = [
-            (1, 17, 2), (1, 17, 10), (1, 18, 0), (1, 18, 2), (2, 16, 1)]
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
-        for version in versions:
-            mock_app.ensure_dependencies.return_value = version
-            with mock.patch(
-                    'quickstart.manage.platform_support.get_juju_command',
-                    side_effect=[self.juju_command]):
-                manage.run(options)
-            mock_app.bootstrap.assert_called_once_with(
-                options.env_name, self.juju_command, requires_sudo=False,
-                debug=options.debug, upload_tools=options.upload_tools,
-                upload_series=options.upload_series,
-                constraints=options.constraints)
-            mock_app.bootstrap.reset_mock()
-
-    def test_local_provider_requiring_sudo(self, mock_app, mock_open):
-        # The application correctly handles working with local providers when
-        # Juju requires an external "sudo" call to bootstrap the environment.
-        # Sudo privileges are required if the Juju version is < 1.17.2.
-        options = self.make_options(env_type='local')
-        versions = [(0, 7, 9), (1, 0, 0), (1, 16, 42), (1, 17, 0), (1, 17, 1)]
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'trusty')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:trusty/juju-gui-42', '0', None, None)
-        for version in versions:
-            mock_app.ensure_dependencies.return_value = version
-            with mock.patch(
-                    'quickstart.manage.platform_support.get_juju_command',
-                    side_effect=[self.juju_command]):
-                manage.run(options)
-            mock_app.bootstrap.assert_called_once_with(
-                options.env_name, self.juju_command, requires_sudo=True,
-                debug=options.debug, upload_tools=options.upload_tools,
-                upload_series=options.upload_series,
-                constraints=options.constraints)
-            mock_app.bootstrap.reset_mock()
-
-    def test_no_local_no_sudo(self, mock_app, mock_open):
-        # Sudo privileges are never required for non-local environments.
-        options = self.make_options(env_type='ec2')
-        mock_app.ensure_dependencies.return_value = (1, 14, 0)
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
-        with mock.patch('quickstart.manage.platform_support.get_juju_command',
-                        side_effect=[self.juju_command]):
+        with self.patch_get_juju_command():
             manage.run(options)
-        mock_app.bootstrap.assert_called_once_with(
-            options.env_name, self.juju_command,
-            requires_sudo=False,
-            debug=options.debug, upload_tools=options.upload_tools,
-            upload_series=options.upload_series,
-            constraints=options.constraints)
 
     def test_no_browser(self, mock_app, mock_open):
         # It is possible to avoid opening the GUI in the browser.
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'trusty')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:trusty/juju-gui-42', '0', None, None)
+        self.configure_app(mock_app, create_auth_token=None)
+        # Run the application.
         options = self.make_options(open_browser=False)
-        manage.run(options)
-        self.assertFalse(mock_open.called)
-
-    def test_admin_secret_fetched(self, mock_app, mock_open):
-        # If an admin secret is fetched from jenv it is used, even if one is
-        # found in environments.yaml, as set in options.admin_secret.
-        mock_app.get_value_from_jenv = self.mock_get_value_from_jenv_success
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
-        options = self.make_options(admin_secret='secret in environments.yaml')
-        manage.run(options)
-        mock_app.connect.assert_has_calls([
-            mock.call(mock_app.get_api_url(), 'from jenv'),
-        ])
-
-    def test_admin_secret_from_environments_yaml(self, mock_app, mock_open):
-        # If an admin secret is not fetched from jenv, then the one from
-        # environments.yaml is used, as found in options.admin_secret.
-        mock_app.get_value_from_jenv = self.mock_get_value_from_jenv_error
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
-        options = self.make_options(admin_secret='secret in environments.yaml')
-        manage.run(options)
-        mock_app.connect.assert_has_calls([
-            mock.call(mock_app.get_api_url(), 'secret in environments.yaml'),
-        ])
-
-    def test_no_admin_secret_found(self, mock_app, mock_open):
-        # If admin-secret cannot be found anywhere a ProgramExit is called.
-        mock_app.ProgramExit = app.ProgramExit
-        mock_app.get_value_from_jenv = self.mock_get_value_from_jenv_error
-        # Make mock_app.bootstrap return the already_bootstrapped flag and the
-        # bootstrap node series.
-        mock_app.bootstrap.return_value = (True, 'precise')
-        # Make mock_app.check_environment return the charm URL, the machine
-        # where to deploy the charm, the service and unit data.
-        mock_app.check_environment.return_value = (
-            'cs:precise/juju-gui-42', '0', None, None)
-        options = self.make_options(
-            admin_secret=None,
-            env_name='local',
-            env_file='environments.yaml')
-        with self.assertRaises(app.ProgramExit) as context:
+        with self.patch_get_juju_command():
             manage.run(options)
-        expected = (
-            u'admin-secret not found in {}/environments/local.jenv '
-            'or environments.yaml'.format(settings.JUJU_HOME))
-        self.assertEqual(expected, context.exception.message)
+        # The browser is not opened.
+        self.assertFalse(mock_open.called)
