@@ -19,14 +19,11 @@
 from __future__ import unicode_literals
 
 import datetime
-import httplib
 import json
 import os
 import shutil
-import socket
 import tempfile
 import unittest
-import urllib2
 
 import mock
 import yaml
@@ -305,38 +302,6 @@ class TestConvertBundleUrl(helpers.ValueErrorTestsMixin, unittest.TestCase):
                 utils.convert_bundle_url(url)
 
 
-class TestGetCharmUrl(helpers.UrlReadTestsMixin, unittest.TestCase):
-
-    def test_charm_url(self):
-        # The Juju GUI charm URL is correctly returned.
-        contents = json.dumps({'charm': {'url': 'cs:trusty/juju-gui-42'}})
-        with self.patch_urlread(contents=contents) as mock_urlread:
-            charm_url = utils.get_charm_url('trusty')
-        self.assertEqual('cs:trusty/juju-gui-42', charm_url)
-        mock_urlread.assert_called_once_with(
-            'http://manage.jujucharms.com/api/3/charm/trusty/juju-gui')
-
-    def test_io_error(self):
-        # IOErrors are properly propagated.
-        with self.patch_urlread(error=True) as mock_urlread:
-            with self.assertRaises(IOError) as context_manager:
-                utils.get_charm_url('precise')
-        mock_urlread.assert_called_once_with(
-            'http://manage.jujucharms.com/api/3/charm/precise/juju-gui')
-        self.assertEqual('bad wolf', bytes(context_manager.exception))
-
-    def test_value_error(self):
-        # A ValueError is raised if the API response is not valid.
-        contents = json.dumps({'charm': {}})
-        with self.patch_urlread(contents=contents) as mock_urlread:
-            with self.assertRaises(ValueError) as context_manager:
-                utils.get_charm_url('trusty')
-        mock_urlread.assert_called_once_with(
-            'http://manage.jujucharms.com/api/3/charm/trusty/juju-gui')
-        self.assertEqual(
-            'unable to find the charm URL', bytes(context_manager.exception))
-
-
 class TestGetQuickstartBanner(unittest.TestCase):
 
     def patch_datetime(self):
@@ -686,72 +651,6 @@ class TestRunOnce(unittest.TestCase):
         self.assertEqual(list.append.__doc__, self.func.__doc__)
 
 
-class TestUrlread(unittest.TestCase):
-
-    def patch_urlopen(self, contents=None, error=None, content_type=None):
-        """Patch the urllib2.urlopen function.
-
-        If contents is not None, the read() method of the returned mock object
-        returns the given contents.
-        If content_type is provided, the response includes the content type.
-        If an error is provided, the call raises the error.
-        """
-        mock_urlopen = mock.MagicMock()
-        if contents is not None:
-            mock_urlopen().read.return_value = contents
-        if content_type is not None:
-            mock_urlopen().headers = {'content-type': content_type}
-        if error is not None:
-            mock_urlopen.side_effect = error
-        mock_urlopen.reset_mock()
-        return mock.patch('urllib2.urlopen', mock_urlopen)
-
-    def test_contents(self):
-        # The URL contents are correctly returned.
-        with self.patch_urlopen(contents=b'URL contents') as mock_urlopen:
-            contents = utils.urlread('http://example.com/path/')
-        self.assertEqual('URL contents', contents)
-        self.assertIsInstance(contents, unicode)
-        mock_urlopen.assert_called_once_with('http://example.com/path/')
-
-    def test_content_type(self):
-        # The URL contents are decoded using the site charset.
-        patch_urlopen = self.patch_urlopen(
-            contents=b'URL contents: \xf8',  # This is not a UTF-8 byte string.
-            content_type='text/html; charset=ISO-8859-1')
-        with patch_urlopen as mock_urlopen:
-            contents = utils.urlread('http://example.com/path/')
-        self.assertEqual('URL contents: \xf8', contents)
-        self.assertIsInstance(contents, unicode)
-        mock_urlopen.assert_called_once_with('http://example.com/path/')
-
-    def test_no_content_type(self):
-        # The URL contents are decoded with UTF-8 by default.
-        patch_urlopen = self.patch_urlopen(
-            contents=b'URL contents: \xf8',  # This is not a UTF-8 byte string.
-            content_type='text/html')
-        with patch_urlopen as mock_urlopen:
-            contents = utils.urlread('http://example.com/path/')
-        self.assertEqual('URL contents: ', contents)
-        self.assertIsInstance(contents, unicode)
-        mock_urlopen.assert_called_once_with('http://example.com/path/')
-
-    def test_errors(self):
-        # An IOError is raised if an error occurs connecting to the API.
-        errors = {
-            'httplib HTTPException': httplib.HTTPException,
-            'socket error': socket.error,
-            'urllib2 URLError': urllib2.URLError,
-        }
-        for message, exception_class in errors.items():
-            exception = exception_class(message)
-            with self.patch_urlopen(error=exception) as mock_urlopen:
-                with self.assertRaises(IOError) as context_manager:
-                    utils.urlread('http://example.com/path/')
-            mock_urlopen.assert_called_once_with('http://example.com/path/')
-            self.assertEqual(message, bytes(context_manager.exception))
-
-
 class TestGetJujuVersion(
         helpers.CallTestsMixin, helpers.ValueErrorTestsMixin,
         unittest.TestCase):
@@ -768,8 +667,14 @@ class TestGetJujuVersion(
             with self.assert_value_error('bad wolf'):
                 utils.get_juju_version('juju')
 
+    def test_alpha_version_string(self):
+        # Patch level defaults to zero.
+        with self.patch_call(0, '1.17-alpha1-precise-amd64', ''):
+            version = utils.get_juju_version('juju')
+        self.assertEqual((1, 17, 0), version)
+
     def test_invalid_version_string(self):
         # A ValueError is raised if "juju version" outputs an invalid version.
-        with self.patch_call(0, '1.17-precise-amd64', ''):
-            with self.assert_value_error('invalid version string: 1.17'):
+        with self.patch_call(0, '1-precise-amd64', ''):
+            with self.assert_value_error('invalid version string: 1'):
                 utils.get_juju_version('juju')
