@@ -63,23 +63,29 @@ container_addresses = [
 
 class TestRetrievePublicAddress(unittest.TestCase):
 
+    def resolver(self, hostname):
+        """A fake resolver returning no errors."""
+        return None
+
     def test_empty_addresses(self):
         # None is returned if there are no available addresses.
-        self.assertIsNone(watchers.retrieve_public_adddress([]))
+        address = watchers.retrieve_public_adddress([], self.resolver)
+        self.assertIsNone(address)
 
     def test_cloud_address_not_found(self):
         # None is returned if a cloud machine public address is not available.
         addresses = [
             {'NetworkName': '',
-             'Scope': 'local-cloud',
-             'Type': 'hostname',
-             'Value': 'eu-west-1.example.internal'},
+             'Scope': 'public',
+             'Type': 'ipv6',
+             'Value': 'fe80::92b8:d0ff:fe94:8f8c'},
             {'NetworkName': '',
              'Scope': 'local-cloud',
-             'Type': 'ipv4',
-             'Value': '10.42.47.10'},
+             'Type': 'ipv6',
+             'Value': 'fe80::216:3eff:fefd:787e'},
         ]
-        self.assertIsNone(watchers.retrieve_public_adddress(addresses))
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertIsNone(address)
 
     def test_container_address_not_found(self):
         # None is returned if an LXC public address is not available.
@@ -89,7 +95,8 @@ class TestRetrievePublicAddress(unittest.TestCase):
             'Type': 'ipv6',
             'Value': 'fe80::216:3eff:fefd:787e',
         }]
-        self.assertIsNone(watchers.retrieve_public_adddress(addresses))
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertIsNone(address)
 
     def test_empty_public_address(self):
         # None is returned if the public address has no value.
@@ -103,17 +110,20 @@ class TestRetrievePublicAddress(unittest.TestCase):
              'Type': 'ipv4',
              'Value': ''},
         ]
-        self.assertIsNone(watchers.retrieve_public_adddress(addresses))
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertIsNone(address)
 
     def test_cloud_addresses(self):
         # The public address of a cloud machine is properly returned.
-        public_address = watchers.retrieve_public_adddress(cloud_addresses)
-        self.assertEqual('eu-west-1.compute.example.com', public_address)
+        address = watchers.retrieve_public_adddress(
+            cloud_addresses, self.resolver)
+        self.assertEqual('eu-west-1.compute.example.com', address)
 
     def test_container_addresses(self):
         # The public address of an LXC instance is properly returned.
-        public_address = watchers.retrieve_public_adddress(container_addresses)
-        self.assertEqual('10.0.3.42', public_address)
+        address = watchers.retrieve_public_adddress(
+            container_addresses, self.resolver)
+        self.assertEqual('10.0.3.42', address)
 
     def test_old_juju_version(self):
         # The public address is properly returned when using Juju < 1.20.
@@ -129,11 +139,32 @@ class TestRetrievePublicAddress(unittest.TestCase):
              'Type': 'hostname',
              'Value': 'eu-west-1.example.internal'},
         ]
-        public_address = watchers.retrieve_public_adddress(addresses)
-        self.assertEqual('eu-west-1.compute.example.com', public_address)
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertEqual('eu-west-1.compute.example.com', address)
 
-    def test_last_unknown_address(self):
-        # If the scope of multiple addresses is unknown, the last one is taken.
+    def test_local_cloud_address(self):
+        # If there are no available public addresses the first ipv4 address
+        # with cloud-local scope is returned.
+        addresses = [
+            {'NetworkName': '',
+             'Scope': 'local-cloud',
+             'Type': 'hostname',
+             'Value': 'eu-west-1.example.internal'},
+            {'NetworkName': '',
+             'Scope': 'local-cloud',
+             'Type': 'ipv4',
+             'Value': '10.42.47.10'},
+            {'NetworkName': '',
+             'Scope': 'local-cloud',
+             'Type': 'ipv4',
+             'Value': '1.2.3.4'},
+        ]
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertEqual('10.42.47.10', address)
+
+    def test_unknown_address(self):
+        # If there are no available public or local-cloud addresses, the first
+        # address with unknown scope is returned.
         addresses = [
             {'NetworkName': '',
              'Scope': '',
@@ -144,8 +175,52 @@ class TestRetrievePublicAddress(unittest.TestCase):
              'Type': 'ipv4',
              'Value': '10.0.3.47'},
         ]
-        public_address = watchers.retrieve_public_adddress(addresses)
-        self.assertEqual('10.0.3.47', public_address)
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertEqual('10.0.3.42', address)
+
+    def test_preferred_fallback_address(self):
+        # If there are no available public addresses the first fallback address
+        # in the list is returned.
+        addresses = [
+            {'NetworkName': '',
+             'Scope': '',
+             'Type': 'ipv4',
+             'Value': '10.0.3.47'},
+            {'NetworkName': '',
+             'Scope': 'local-cloud',
+             'Type': 'ipv4',
+             'Value': '10.42.47.10'},
+
+        ]
+        address = watchers.retrieve_public_adddress(addresses, self.resolver)
+        self.assertEqual('10.0.3.47', address)
+        # Now test with a reversed order.
+        address = watchers.retrieve_public_adddress(
+            reversed(addresses), self.resolver)
+        self.assertEqual('10.42.47.10', address)
+
+    def test_unresolvable_public_address(self):
+        # None is returned if a public cloud is found but it is not resolvable.
+        addresses = [
+            {'NetworkName': '',
+             'Scope': 'public',
+             'Type': 'hostname',
+             'Value': 'eu-west-1.example.internal'},
+        ]
+        expected_warning = (
+            'cannot resolve public eu-west-1.example.internal address, '
+            'looking for another candidate: bad wolf')
+        with helpers.assert_logs([expected_warning], level='warn'):
+            address = watchers.retrieve_public_adddress(
+                addresses, lambda hostname: 'bad wolf')
+        self.assertIsNone(address)
+
+    def test_unresolvable_public_address_fallback(self):
+        # If there are no resolvable public addresses the first ipv4 address
+        # with cloud-local scope is returned.
+        address = watchers.retrieve_public_adddress(
+            cloud_addresses, lambda hostname: 'bad wolf')
+        self.assertEqual('10.42.47.10', address)
 
 
 class TestParseMachineChange(helpers.ValueErrorTestsMixin, unittest.TestCase):
@@ -205,8 +280,12 @@ class TestParseMachineChange(helpers.ValueErrorTestsMixin, unittest.TestCase):
         # A message is printed to stdout when the machine obtains a public
         # address.
         data = {'Addresses': cloud_addresses, 'Id': '1', 'Status': 'pending'}
-        status, address = watchers.parse_machine_change(
-            'change', data, 'pending', '')
+        # Patch the hostname resolver used to check hostname addresses.
+        # Note that resolve error cases are properly exercised by the
+        # watchers.retrieve_public_adddress tests.
+        with helpers.patch_check_resolvable():
+            status, address = watchers.parse_machine_change(
+                'change', data, 'pending', '')
         self.assertEqual('pending', status)
         self.assertEqual('eu-west-1.compute.example.com', address)
         mock_print.assert_called_once_with(
@@ -248,8 +327,8 @@ class TestParseUnitChange(helpers.ValueErrorTestsMixin, unittest.TestCase):
         # A ValueError is raised if the change represents a unit removal.
         data = {'Name': 'django/42', 'Status': 'started'}
         with self.assert_value_error('django/42 unexpectedly removed'):
-            # The last two arguments are the current status and address.
-            watchers.parse_unit_change('remove', data, '', '')
+            # The last argument is the current status.
+            watchers.parse_unit_change('remove', data, '')
 
     def test_unit_error(self):
         # A ValueError is raised if the unit is in an error state.
@@ -260,120 +339,55 @@ class TestParseUnitChange(helpers.ValueErrorTestsMixin, unittest.TestCase):
         }
         expected_error = 'django/0 is in an error state: start error: bad wolf'
         with self.assert_value_error(expected_error):
-            # The last two arguments are the current status and address.
-            watchers.parse_unit_change('change', data, '', '')
-
-    @helpers.mock_print
-    def test_address_notified(self, mock_print):
-        # A message is printed to stdout when the unit obtains a public
-        # address. The function returns the status, the new address and the
-        # machine identifier.
-        data = {
-            'Name': 'haproxy/2',
-            'Status': 'pending',
-            'PublicAddress': 'haproxy2.example.com',
-            'MachineId': '42',
-        }
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, 'pending', '')
-        self.assertEqual('pending', status)
-        self.assertEqual('haproxy2.example.com', address)
-        self.assertEqual('42', machine_id)
-        mock_print.assert_called_once_with(
-            'haproxy/2 placed on haproxy2.example.com')
+            # The last argument is the current status.
+            watchers.parse_unit_change('change', data, '')
 
     @helpers.mock_print
     def test_pending_status_notified(self, mock_print):
         # A message is printed to stdout when the unit changes its status to
-        # "pending". The function returns the new status, the address and the
-        # machine identifier. The last two values are empty strings if the unit
-        # has not yet been assigned to a machine.
-        data = {'Name': 'django/1', 'Status': 'pending', 'PublicAddress': ''}
-        # The last two arguments are the current status and address.
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, '', '')
+        # "pending". The function returns the new status and the machine
+        # identifier, which is empty if the unit has not yet been assigned to a
+        # machine.
+        data = {'Name': 'django/1', 'Status': 'pending'}
+        # The last argument is the current status.
+        status, machine_id = watchers.parse_unit_change('change', data, '')
         self.assertEqual('pending', status)
-        self.assertEqual('', address)
         self.assertEqual('', machine_id)
         mock_print.assert_called_once_with('django/1 deployment is pending')
 
     @helpers.mock_print
     def test_installed_status_notified(self, mock_print):
         # A message is printed to stdout when the unit changes its status to
-        # "installed". The function returns the new status, the address and the
-        # machine identifier.
-        data = {
-            'Name': 'django/42',
-            'Status': 'installed',
-            'PublicAddress': 'django42.example.com',
-            'MachineId': '1',
-        }
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, 'pending', 'django42.example.com')
+        # "installed".
+        data = {'Name': 'django/42', 'Status': 'installed', 'MachineId': '1'}
+        status, machine_id = watchers.parse_unit_change(
+            'change', data, 'pending')
         self.assertEqual('installed', status)
-        self.assertEqual('django42.example.com', address)
         self.assertEqual('1', machine_id)
         mock_print.assert_called_once_with('django/42 is installed')
 
     @helpers.mock_print
     def test_started_status_notified(self, mock_print):
         # A message is printed to stdout when the unit changes its status to
-        # "started". The function returns the new status, the address and the
-        # machine identifier.
-        data = {
-            'Name': 'wordpress/0',
-            'Status': 'started',
-            'PublicAddress': 'wordpress0.example.com',
-            'MachineId': '0',
-        }
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, '', 'wordpress0.example.com')
+        # "started".
+        data = {'Name': 'wordpress/0', 'Status': 'started', 'MachineId': '0'}
+        # The last argument is the current status.
+        status, machine_id = watchers.parse_unit_change('change', data, '')
         self.assertEqual('started', status)
-        self.assertEqual('wordpress0.example.com', address)
         self.assertEqual('0', machine_id)
         mock_print.assert_called_once_with('wordpress/0 is ready on machine 0')
 
-    @helpers.mock_print
-    def test_both_status_and_address_notified(self, mock_print):
-        # Both status and public address changes are notified if required.
-        data = {
-            'Name': 'django/0',
-            'Status': 'started',
-            'PublicAddress': 'django42.example.com',
-            'MachineId': '0',
-        }
-        # The last two arguments are the current status and address.
-        watchers.parse_unit_change('change', data, '', '')
-        self.assertEqual(2, mock_print.call_count)
-        mock_print.assert_has_calls([
-            mock.call('django/0 placed on django42.example.com'),
-            mock.call('django/0 is ready on machine 0'),
-        ])
-
     @helpers.mock_print
     def test_status_not_changed(self, mock_print):
         # If the status in the unit change and the given current status are the
         # same value, nothing is printed and the current values are returned.
-        data = {'Name': 'django/1', 'Status': 'pending', 'PublicAddress': ''}
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, 'pending', '')
+        data = {'Name': 'django/1', 'Status': 'pending'}
+        status, machine_id = watchers.parse_unit_change(
+            'change', data, 'pending')
         self.assertEqual('pending', status)
-        self.assertEqual('', address)
         self.assertEqual('', machine_id)
         self.assertFalse(mock_print.called)
 
-    @helpers.mock_print
-    def test_address_not_available(self, mock_print):
-        # An empty address is returned when the public address field is not
-        # included in the change data.
-        data = {'Name': 'haproxy/2', 'Status': 'pending', 'MachineId': '42'}
-        status, address, machine_id = watchers.parse_unit_change(
-            'change', data, 'pending', '')
-        self.assertEqual('pending', status)
-        self.assertEqual('', address)
-        self.assertEqual('42', machine_id)
-        self.assertFalse(mock_print.called)
-
 
 class TestUnitMachineChanges(unittest.TestCase):
 
