@@ -25,15 +25,11 @@ import collections
 import datetime
 import errno
 import functools
-import httplib
-import json
 import logging
 import os
 import pipes
 import re
-import socket
 import subprocess
-import urllib2
 
 import quickstart
 from quickstart import (
@@ -150,21 +146,6 @@ def convert_bundle_url(bundle_url):
             bundle_id)
 
 
-def get_charm_url(series):
-    """Return the charm URL of the latest Juju GUI charm revision.
-
-    Raise an IOError if any problems occur connecting to the API endpoint.
-    Raise a ValueError if the API returns invalid data.
-    """
-    url = settings.CHARMWORLD_API.format(
-        series=series, charm=settings.JUJU_GUI_CHARM_NAME)
-    charm_info = json.loads(urlread(url))
-    charm_url = charm_info.get('charm', {}).get('url')
-    if charm_url is None:
-        raise ValueError(b'unable to find the charm URL')
-    return charm_url
-
-
 def get_quickstart_banner():
     """Return a quickstart banner suitable for being included in files.
 
@@ -344,6 +325,8 @@ def get_juju_version(juju_command):
     Return a (major:int, minor:int, patch:int) tuple, including major, minor
     and patch version numbers.
 
+    Handle the special case of no patch level by defaulting to 0.
+
     Raise a ValueError if the "juju version" call exits with an error
     or the returned version is not well formed.
     """
@@ -352,8 +335,11 @@ def get_juju_version(juju_command):
         raise ValueError(error.encode('utf-8'))
     version_string = output.split('-')[0]
     try:
-        major, minor, patch = version_string.split('.', 2)
-        return int(major), int(minor), int(patch)
+        parts = version_string.split('.', 2)
+        if len(parts) == 2:
+            parts.append(0)
+        major, minor, patch = map(int, parts)
+        return major, minor, patch
     except ValueError:
         msg = 'invalid version string: {}'.format(version_string)
         raise ValueError(msg.encode('utf-8'))
@@ -371,24 +357,3 @@ def run_once(function):
             return function(*args, **kwargs)
     decorated.called = False
     return decorated
-
-
-def urlread(url):
-    """Open the given URL and return the page contents.
-
-    Raise an IOError if any problems occur.
-    """
-    try:
-        response = urllib2.urlopen(url)
-    except urllib2.URLError as err:
-        raise IOError(err.reason)
-    except (httplib.HTTPException, socket.error, urllib2.HTTPError) as err:
-        raise IOError(bytes(err))
-    contents = response.read()
-    content_type = response.headers['content-type']
-    charset = 'utf-8'
-    if 'charset=' in content_type:
-        sent_charset = content_type.split('charset=')[-1].strip()
-        if sent_charset:
-            charset = sent_charset
-    return contents.decode(charset, 'ignore')
