@@ -21,15 +21,25 @@ from __future__ import (
     unicode_literals,
 )
 
+import logging
 
+from quickstart import netutils
+
+
+IPV4_ADDRESS = 'ipv4'
 IPV6_ADDRESS = 'ipv6'
-NETWORK_PUBLIC = 'public'
-NETWORK_UNKNOWN = ''
+SCOPE_CLOUD_LOCAL = 'local-cloud'
+SCOPE_PUBLIC = 'public'
+SCOPE_UNKNOWN = ''
 
 
-def retrieve_public_adddress(addresses):
+def retrieve_public_adddress(addresses, hostname_resolver):
     """Parse the given addresses and return a public address if available.
 
+    Use the given hostname_resolver callable to ensure a candidate address can
+    be resolved. The given function must accept an hostname/ip address and
+    return None if the address is resolvable, or an error string otherwise.
+
     The addresses argument is a list of address dictionaries.
     Cloud addresses look like the following:
 
@@ -73,8 +83,8 @@ def retrieve_public_adddress(addresses):
     found, this function returns None.
     """
     # This implementation reflects how the public address is retrieved in Juju:
-    # see juju-core/instance/address.go:SelectPublicAddress.
-    public_address = None
+    # see juju-core/network/address.go:SelectPublicAddress.
+    fallback_address = None
     for address in addresses:
         value = address['Value']
         # Exclude empty values and ipv6 addresses.
@@ -82,14 +92,33 @@ def retrieve_public_adddress(addresses):
             # The address scope was called "NetworkScope" prior to juju 1.20.0.
             scope = address.get('Scope', address.get('NetworkScope'))
             # If the scope is public then we have found the address.
-            if scope == NETWORK_PUBLIC:
-                return value
-            # If the scope is unknown then store the value. This way the last
-            # address with unknown scope will be returned, and we are able to
-            # return the right LXC address.
-            if scope == NETWORK_UNKNOWN:
-                public_address = value
-    return public_address
+            if scope == SCOPE_PUBLIC:
+                error = hostname_resolver(value)
+                if error is None:
+                    return value
+                # If the address is not resolvable, fall back to the next
+                # candidate, whether it is another public address, a
+                # local-cloud one or an unknown one. The local-cloud case, for
+                # instance, would happen when using the maas provider and the
+                # maas DNS server is not configured locally.
+                logging.warn(
+                    'cannot resolve public {} address, looking for another '
+                    'candidate: {}'.format(value, error))
+            if fallback_address is not None:
+                # If we already have a fallback address then we can jump to the
+                # next candidate.
+                continue
+            address_type = address.get('Type')
+            if scope == SCOPE_CLOUD_LOCAL and address_type == IPV4_ADDRESS:
+                # Use the local cloud scoped address as fallback value: the
+                # current environment is probably maas and the public address
+                # of the instance is not resolvable by local DNS.
+                fallback_address = value
+            elif scope == SCOPE_UNKNOWN:
+                # Use the unknown scoped address as fallback value: the current
+                # environment is probably local.
+                fallback_address = value
+    return fallback_address
 
 
 def parse_machine_change(action, data, current_status, address):
@@ -120,7 +149,8 @@ def parse_machine_change(action, data, current_status, address):
     # of units hosted by a specific machine.
     if not address:
         addresses = data.get('Addresses', [])
-        public_address = retrieve_public_adddress(addresses)
+        public_address = retrieve_public_adddress(
+            addresses, netutils.check_resolvable)
         if public_address is not None:
             address = public_address
             print('unit placed on {}'.format(address))
@@ -134,16 +164,16 @@ def parse_machine_change(action, data, current_status, address):
     return status, address
 
 
-def parse_unit_change(action, data, current_status, address):
+def parse_unit_change(action, data, current_status):
     """Parse the given unit change.
 
     The change is represented by the given action/data pair.
-    Also receive the last known unit status and address, which can be empty
-    strings if those pieces of information are unknown.
+    Also receive the last known unit status which can be empty if the unit
+    status is not yet known.
 
     Output a human readable message each time a relevant change is found.
 
-    Return the unit status, address and machine identifier.
+    Return the unit status and the identifier of the machine hosting the unit.
     Raise a ValueError if the service unit is removed or in an error state.
     """
     unit_name = data['Name']
@@ -157,15 +187,6 @@ def parse_unit_change(action, data, current_status, address):
         msg = '{} is in an error state: {}: {}'.format(
             unit_name, status, data['StatusInfo'])
         raise ValueError(msg.encode('utf-8'))
-    # Notify when the unit becomes reachable. Up to juju-core 1.18, the
-    # mega-watcher for units includes the public address for each unit. This
-    # info is likely to be deprecated in favor of addresses as included in the
-    # mega-watcher for machines, but we still try to retrieve the address here
-    # for backward compatibility.
-    if not address:
-        address = data.get('PublicAddress', '')
-        if address:
-            print('{} placed on {}'.format(unit_name, address))
     # Notify status changes.
     if status != current_status:
         if status == 'pending':
@@ -175,7 +196,7 @@ def parse_unit_change(action, data, current_status, address):
         elif status == 'started':
             print('{} is ready on machine {}'.format(
                 unit_name, data['MachineId']))
-    return status, address, data.get('MachineId', '')
+    return status, data.get('MachineId', '')
 
 
 def unit_machine_changes(changeset):
